/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef DSIMPLEX_HPP
#define DSIMPLEX_HPP

#ifndef NOMINMAX
#define NOMINMAX
#endif

#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <limits>
#include <vector>

namespace optimization {

static void sum_parameters(const std::vector<std::vector<double>>& vertices,
                           std::vector<double>& sums);

template <class EvalFunc>
double move_vertex(std::vector<std::vector<double>>& vertices,
                   std::vector<double>& scores, std::vector<double>& sums,
                   std::vector<double>& trialvalues, size_t index, double fac,
                   EvalFunc& eval) {
  const auto nparam = sums.size();
  const auto fac1 = (1.0 - fac) / nparam;
  const auto fac2 = fac1 - fac;
  for (auto j = size_t{0}; j < nparam; ++j)
    trialvalues[j] = sums[j] * fac1 - vertices[index][j] * fac2;
  const auto score = eval(trialvalues);
  if (score < scores[index]) {
    scores[index] = score;
    for (auto j = size_t{0}; j < nparam; ++j) {
      sums[j] += trialvalues[j] - vertices[index][j];
      vertices[index][j] = trialvalues[j];
    }
  }
  return score;
}

// eval must be a callable object that takes a const std::vector<double>&
// parameter and returns a double
template <class InputIter, class OutputIter, class EvalFunc>
void dsimplex(InputIter first, InputIter last, OutputIter result,
              EvalFunc&& eval, double lambda,
              double tolerance = std::numeric_limits<double>::epsilon() * 2.0,
              double max_iterations = 1000) {
  const auto nparam = std::distance(first, last);
  const auto nvertices = nparam + 1;
  auto iterations = 0;
  auto scorerange = tolerance * 2.0;
  auto vertices = std::vector<std::vector<double>>(nvertices);
  auto scores = std::vector<double>(nvertices);
  auto trialvalues = std::vector<double>(nparam);
  auto sums = std::vector<double>(nparam);

  for (auto& v : vertices) {
    std::copy(first, last, std::back_inserter(v));
  }
  for (auto i = 0l; i < nparam; ++i) {
    vertices[i + 1][i] += lambda;
  }
  std::transform(vertices.begin(), vertices.end(), scores.begin(), eval);
  sum_parameters(vertices, sums);
  for (;;) {
    auto low = 0l;
    auto high = 1l;
    auto next_highest = 0l;
    if (scores[0] > scores[1]) {
      high = 0;
      next_highest = 1;
    }
    for (auto i = 0l; i < nvertices; ++i) {
      if (scores[i] <= scores[low])
        low = i;
      if (scores[i] > scores[high]) {
        next_highest = high;
        high = i;
      } else if (scores[i] > scores[next_highest] && i != high)
        next_highest = i;
    }
    scorerange = 2.0 * fabs(scores[high] - scores[low]) /
                 (fabs(scores[high]) + fabs(scores[low]) +
                  std::numeric_limits<double>::min());
    if (scorerange < tolerance || iterations >= max_iterations) {
      std::copy(vertices[low].cbegin(), vertices[low].cend(), result);
      break;
    } else {
      auto score =
          move_vertex(vertices, scores, sums, trialvalues, high, -1.0, eval);
      if (score <= scores[low])
        move_vertex(vertices, scores, sums, trialvalues, high, 2.0, eval);
      else if (score >= scores[next_highest]) {
        auto savedscore = scores[high];
        score =
            move_vertex(vertices, scores, sums, trialvalues, high, 0.5, eval);
        if (score >= savedscore) {
          for (auto i = 0l; i < nvertices; ++i) {
            if (i == low)
              continue;
            for (auto j = 0l; j < nparam; ++j) {
              vertices[i][j] = 0.5 * (vertices[i][j] + vertices[low][j]);
            }
            scores[i] = eval(vertices[i]);
          }
          sum_parameters(vertices, sums);
        }
      }
    }
    ++iterations;
  }
}

void sum_parameters(const std::vector<std::vector<double>>& param_vec,
                    std::vector<double>& sums) {
  const auto nparam = sums.size();
  for (auto i = size_t{0}; i < nparam; ++i) {
    sums[i] = 0.0;
    for (const auto& p : param_vec) {
      sums[i] += p[i];
    }
  }
}
}

#endif
