/*
 Copyright (c) 2015-2016 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef NOMINMAX
#define NOMINMAX
#endif

#include "epmr.hpp"

#include "dsimplex.hpp"
#include "evolutionaryoptimizer.hpp"
#include "geometry.hpp"
#include "mtz.hpp"
#include "platform.hpp"
#include "pdb.hpp"
#include "reflection.hpp"
#include "rfactor.hpp"
#include "scorer.hpp"
#include "spacegroup.hpp"
#include "statistics.hpp"

#include <algorithm>
#include <cfloat>
#include <cmath>
#include <complex>
#include <cstdio>
#include <fstream>
#include <iomanip>
#include <iterator>
#include <stdexcept>

using std::string;
using std::unordered_map;
using std::unique_ptr;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ios_base;
using std::endl;
using std::runtime_error;
using std::numeric_limits;
using std::flush;
using std::fixed;
using std::setw;
using std::showpoint;
using std::setprecision;
using std::to_string;

using pdb::operator>>;
using namespace crystallography;

const unordered_map<string, string> epmr_defaults = {

    {"version", "16.07.1"},

    // output options
    {"output.level", "all"}, // "best", "none"
    {"output.prefix", "epmr"},
    {"output.reassign-chain-ids", "true"}, // only when copies > 1 or assembly
    {"report.stats", "none"},              // "running", "final"
    {"report.header", "true"},
    {"report.inputs", "true"},
    {"report.options", "true"},
    {"report.steps", "true"},
    {"report.time.total", "true"},
    {"report.evolution", "progress"}, // "generations", "none"
    {"report.local", "true"},
    {"report.solution", "true"},
    {"report.solution.normalized", "true"},

    // file paths
    {"cell.source", ""},
    {"model.source", "-"}, // cin
    {"data.source", "-"},  // cin
    {"data.mtz-column", ""},
    // static structure
    {"structure.source", ""},

    // resolution limits
    {"resolution.low", "80.0"},
    {"resolution.high", "3.0"},

    // molecular transform
    {"transform.scale", "2.0"},
    {"transform.fft.sampling", "3.0"},

    // general options
    {"attempts", "20"}, // number of runs to attempt
    {"copies", "1"},    // total number of copies of the search model to find
    {"pointgroupsearch", "false"},
    {"assembly", "false"}, // true = treat multiple models
                           // as substructures rather than alternatives

    // optimization options
    {"optimizer.global", "evolutionary"}, // "none"
    {"optimizer.local", "simplex"},       // ["powell"], "none"
    {"optimizer.goal", "maximum"},        // "maximum"

    // optimizable parameters
    {"optimize.orientation", "true"},
    {"optimize.position", "true"},
    {"optimize.model", "false"}, // make the model ID an optimizable parameter
    {"optimize.spacegroup",
     "false"}, // make the space group an optimizable parameter
    {"optimize.copies",
     "false"}, // search for all copies of the search model simultaneously
    {"optimize.assembly",
     "false"}, // search with all models in assembly simultaneously
    {"optimize.babinet", "true"},
    {"optimize.local-segments", "true"},

    // scoring
    {"target.type", "ccf"},
    {"target.global.interpolate-fc", "true"},
    //    { "target.local.interpolate-fc", "true" },
    //    { "target.reflections", "all" }, // "random"
    {"target.sigma-weighting", "false"},

    // evolutionary optimizer options
    {"evolution.generations", "50"},
    {"evolution.population", "300"},
    //    { "evolution.competitors", "0" }, // 0=deterministic selection
    {"evolution.selection.rate.starting", "0.4"},     // 0->1
    {"evolution.selection.rate.ending", "0.01"},      // 0->1 // 0.05
    {"evolution.mutation.crossover.rate", "0.6"},     // 0->1
    {"evolution.mutation.interpolation.rate", "0.6"}, // 0->1
    {"evolution.mutation.rotation.rate", "0.5"},      // 0->1 // 0.2
    {"evolution.mutation.rotation.size", "0.05"},     // 0->1
    {"evolution.mutation.rotation.type", "gaussian"},
    {"evolution.mutation.rotation.notrans.rate", "0.9"}, // 0->1 // 0.2
    {"evolution.mutation.rotation.notrans.size", "0.3"}, // 0->1
    {"evolution.mutation.translation.rate", "0.7"},      // 0->1
    {"evolution.mutation.translation.size", "0.2"},      // 0->1
    {"evolution.mutation.translation.type", "gaussian"},

    // When there are multiple segments, only the first segment is affected by
    // the following translation options.
    {"evolution.mutation.translation.minimum.x", "0.0"},
    {"evolution.mutation.translation.minimum.y", "0.0"},
    {"evolution.mutation.translation.minimum.z", "0.0"},
    {"evolution.mutation.translation.range.x", "1.0"}, // 0->1
    {"evolution.mutation.translation.range.y", "1.0"}, // 0->1
    {"evolution.mutation.translation.range.z", "1.0"}, // 0->1
    {"evolution.mutation.modelswap.rate", "0.25"},     // 0->1
    {"evolution.mutation.cellswap.rate", "0.25"},      // 0->1
    // Increasing the babinet mutation rate dramatically improves the
    // success rate for some cases but has a negative effect for most.
    {"evolution.mutation.babinet.rate", "0.1"}, // 0->1
    {"evolution.mutation.babinet.k.size", "0.1"},
    {"evolution.mutation.babinet.B.size", "20.0"},
    {"evolution.mutation.babinet.k.starting", "0.8"},
    {"evolution.mutation.babinet.k.min", "0.2"},
    {"evolution.mutation.babinet.k.max", "1.5"},
    {"evolution.mutation.babinet.B.starting", "200.0"},
    {"evolution.mutation.babinet.B.min", "100.0"},
    {"evolution.mutation.babinet.B.max", "400.0"},

    // downhill simplex optimizer options
    {"simplex.tolerance", "0.000001"},
    {"simplex.size", "0.05"},

    // random number generation
    {"random.seed", "0"}, // 0 = use system-generated seed

    // packing
    // final packing clash measure
    {"packing.clash.type", "interatomic"}, // "interatomic"
    {"packing.clash.distance", "3.4"},
    {"packing.clash.atoms", "c-alpha"},
    // threshold to apply penalty to score (simple center-center packing
    // distance)
    //{ "bump.threshold", "0.0" },
};

static void int_to_sf_simple(struct reflection& ref);
static vector<unit_cell> load_cells(const string& path);
static void load_models(const string& path, vector<vector<atom>>& models);
static void load_model(istream& atomstream, vector<vector<atom>>& models,
                       const string& path);
static vector<reflection> load_reflections(const unit_cell& cell,
                                           double low_res_limit,
                                           double high_res_limit,
                                           const string& path,
                                           string& mtz_refl_column);
static void create_alternate_cells(vector<unit_cell>& cells,
                                   vector<unsigned int>& permutations);
static void canonicalize(std::array<double, 3>& position);
static void
reconcile_options(unordered_map<string, string>& options,
                  const unordered_map<string, string>& user_options);
static void
calc_transforms(const vector<vector<atom>>& models, size_t firstindex, size_t n,
                double highres_limit, double scale, double sampling,
                vector<unique_ptr<molecular_transform<float>>>& transforms);
static bool split_segments(solution& sol, vector<vector<atom>>& modelvec,
                           vector<unit_cell>& cells,
                           vector<std::array<double, 3>>& centers);
static void calc_static(const vector<vector<atom>>& structure,
                        const unit_cell& cell, unsigned int hkl_permutation,
                        vector<reflection>& ref, double res_limit,
                        double sampling);
static void calc_segment_transforms(
    const solution& sol,
    vector<unique_ptr<molecular_transform<float>>>& transforms,
    const vector<vector<atom>>& models, double highres, double scale,
    double sampling);
static void append_structure(const solution& sol,
                             vector<vector<atom>>& structure,
                             const vector<vector<atom>>& models,
                             const vector<unit_cell>& cells,
                             vector<std::array<double, 3>>& centers);
static int calc_packing_clashes(const solution& sol,
                                const vector<vector<atom>>& models,
                                const vector<vector<atom>>& structure,
                                const vector<unit_cell>& cells,
                                bool use_all_atoms, double clash_distance_sq);
static void print_option_summary(unordered_map<string, string>& options,
                                 ostream& coordstr);
static void fix_polar_axes(std::array<bool, 3>& fixed_axes,
                           const spacegroup& sg);
// static std::array<double, 3> calc_cns_euler_angles(const std::array<double,
// 4>& q);
static std::array<double, 3>
calc_ccp4_euler_angles(const std::array<double, 4>& q);
static void transform_segment_toward_point(rigid_body& sol,
                                           const unit_cell& cell,
                                           const std::array<double, 3>& point,
                                           std::array<double, 3>& fixed_origin);
static void
write_atom_record(const atom& atm,
                  const std::array<std::array<double, 4>, 4>& transform,
                  char chain_id, ostream& coordstr);
static void report_cells(const vector<unit_cell>& cells, ostream& outs);
static void report_solution(const solution& s, ostream& outs, double score,
                            double rfactor, int attempt,
                            vector<unit_cell>& cells, int clashes);
static void report_stats(size_t spcgrp, size_t model, int copy, int attempt,
                         vector<double>& scores, ostream& outs);
static void write_solution(const vector<vector<atom>>& structure,
                           const unit_cell& cell, double score, double rfactor,
                           int clashes, const string& path, ostream& outstr,
                           bool reassign_chains);
static void write_atoms(ostream& coordstream, double score, double rfactor,
                        int clashes, const unit_cell& cell,
                        const vector<vector<atom>>& structure,
                        bool reassign_chains);
static void
normalize_local_search_solution(solution& sol, const vector<unit_cell>& cells,
                                const vector<std::array<double, 3>>& centers,
                                const std::array<bool, 3>& fixed_axes);
static void
normalize_global_search_solution(solution& sol, const vector<unit_cell>& cells,
                                 const vector<std::array<double, 3>>& centers,
                                 const std::array<bool, 3>& fixed_axes,
                                 bool have_partial_str);
static void normalize_solution(solution& sol, const vector<unit_cell>& cells,
                               const vector<std::array<double, 3>>& centers,
                               const std::array<bool, 3>& fixed_axes,
                               bool have_partial_str, bool is_local_search);
static void dsimplex_optimize(solution& sol,
                              scorer<pearson<float>, interpolation::Linear,
                                     local_reflection_parallelism>& scorer,
                              double simplex_size, double simplex_tolerance,
                              bool rotate, bool translate, bool maximizing);

inline static bool istrue(const unordered_map<string, string>& options,
                          const string& key) {
  return options.at(key) == "true";
}

void epmr(const unordered_map<string, string>& user_options, ostream& outs) {

  const auto epmr_start_time = std::chrono::steady_clock::now();

  auto options = epmr_defaults;
  for (const auto& o : user_options) {
    options.at(o.first) = o.second;
  }

  if (istrue(options, "report.header")) {
    print_epmr_header(outs);
  }

  auto report_inputs = istrue(options, "report.inputs");
  auto report_steps = istrue(options, "report.steps");

  const auto& cellfile = options.at("cell.source");
  if (report_inputs) {
    outs << "Reading unit cell and space group information from " << cellfile
         << endl;
  }
  auto cells = load_cells(cellfile);
  auto hkl_permutations = vector<unsigned int>();
  for (auto i = size_t{0}; i < cells.size(); ++i) {
    hkl_permutations.push_back(0);
  }
  if (istrue(options, "pointgroupsearch")) {
    create_alternate_cells(cells, hkl_permutations);
  }
  if (report_steps) {
    report_cells(cells, outs);
  }

  const auto low_res_limit = stod(options.at("resolution.low"));
  const auto high_res_limit = stod(options.at("resolution.high"));
  const auto& hklfile = options.at("data.source");
  auto mtz_refl_col = options.at("data.mtz-column");
  if (report_inputs) {
    outs << "Reading reflections from " << hklfile << endl;
  }
  auto refl = load_reflections(cells[0], low_res_limit, high_res_limit, hklfile,
                               mtz_refl_col);
  // Insert the MTZ column label into the output stream so users
  // have a diagnostic.
  if (report_steps) {
    if (mtz_refl_col.length() > 0) {
      outs << "Used MTZ data column '" + mtz_refl_col << "'\n";
    }
    outs << refl.size() << " reflections in resolution range\n\n";
  }
  if (refl.size() == 0) {
    throw runtime_error("no input reflections in the resolution range");
  }

  if (istrue(options, "target.sigma-weighting")) {
    for (auto& r : refl) {
      r.sigma = 1.0f / (r.sigma * r.sigma);
    }
  } else {
    for (auto& r : refl) {
      r.sigma = 1.0f;
    }
  }

  // assumes all target cells have the same dimensions
  // (permuted axes don't cause a problem)
  for (auto i = size_t{0}; i < refl.size(); ++i) {
    refl[i].stolsq = static_cast<float>(
        cells[0].stolsq(refl[i].hkl[0], refl[i].hkl[1], refl[i].hkl[2]));
  }

  auto models = vector<vector<atom>>(); // search models, cartesian
  auto ortho_centers =
      vector<std::array<double, 3>>(); // model centers, orthogonal
  auto structure =
      vector<vector<atom>>(); // static partial structure, fractional
  auto partial_structure_centers =
      vector<std::array<double, 3>>(); // centers, fractional

  if (report_inputs) {
    auto modelsrc = options.at("model.source");
    if (modelsrc.length() == 0 || modelsrc == "-") {
      modelsrc = "standard input";
    }
    if (modelsrc[0] == '@') {
      modelsrc = modelsrc.substr(1);
      outs << "Loading search models listed in " << modelsrc << endl;
    } else {
      outs << "Loading search model from " << modelsrc << endl;
    }
  }
  load_models(options.at("model.source"), models);
  if (report_steps) {
    for (auto i = size_t{0}; i < models.size(); ++i) {
      outs << "Model " << (i + 1) << ": " << models[i].size() << " atoms"
           << '\n';
    }
    outs << endl;
  }
  if (models.size() == 0) {
    throw runtime_error("no search model provided.");
  }
  for (const auto& m : models) {
    if (m.size() == 0) {
      throw runtime_error("no atoms in search model");
    }
  }
  // translate all models to the origin
  for (auto& atoms : models) {
    auto ave = std::array<double, 3>{};
    for (const auto& a : atoms) {
      ave = geom::add(ave, a.position);
    }
    geom::scale(ave, 1.0 / atoms.size(), ave);
    for (auto& a : atoms) {
      a.position = geom::sub(a.position, ave);
    }
    ortho_centers.emplace_back(ave);
  }

  // load static structure
  if (options.at("structure.source").length() > 0) {

    std::istringstream file_name_stream(options.at("structure.source"));
    auto next_file_name = string{};
    while (file_name_stream >> next_file_name) {
      if (report_inputs)
        outs << "Loading static structure from " << next_file_name << endl;
      load_models(next_file_name, structure);
    }
    if (report_steps) {
      outs << structure.size() << " coordinate file(s) loaded.\n" << endl;
    }

    if (structure.size() == 0)
      throw runtime_error("unable to load static structure.");
    for (auto& atoms : structure) {
      if (atoms.size() == 0) {
        throw runtime_error("no atoms in static-structure model");
      }
      auto coord_ave = std::array<double, 3>{};
      for (atom& a : atoms) {
        a.position = cells[0].fractionalize(a.position);
        coord_ave = geom::add(coord_ave, a.position);
      }
      geom::scale(coord_ave, 1.0 / atoms.size(), coord_ave);
      partial_structure_centers.push_back(coord_ave);
    }
  }

  auto initial_structure_size =
      structure.size(); // number of pieces of partial structure at start

  auto seed = static_cast<std::mt19937_64::result_type>(
      stoul(options.at("random.seed")));
  if (seed == 0) {
    std::random_device randev{};
    seed = randev();
    if (std::mt19937_64::max() > std::random_device::max()) {
      seed = (seed << 32) + randev();
    }
    options.at("random.seed") = to_string(seed);
  }
  auto rng = std::mt19937_64(seed);

  auto maximizing = options.at("optimizer.goal") == "maximum";

  if (models.size() == 1) {
    options.at("optimize.assembly") = "false";
  }
  reconcile_options(options, user_options);
  if (istrue(options, "report.options")) {
    print_option_summary(options, outs);
  }
  if (istrue(options, "report.solution")) {
    outs << "The following information will be reported for each solution:\n"
         << "  CC          linear correlation coefficient between Fo and Fc\n"
         << "  R           crystallographic R-factor\n"
         << "  d<          number of close Calpha-Calpha packing interactions "
            "(<3.4Ang)\n"
         << "  al, be, ga  rotation (Eulerian angles, CCP4 convention)\n"
         << "  x, y, z     translation (fractional, for origin-centered search "
            "model)\n"
         << "  mod         search model ID\n"
         << "  sg          space group number\n\n";
  }
  auto best_scores = vector<vector<vector<double>>>(); // per cell/model/copy
  auto best_rfactors = vector<vector<vector<double>>>();
  auto best_solutions = vector<vector<vector<solution>>>();

  auto transforms = vector<unique_ptr<molecular_transform<float>>>{};

  auto ncel = istrue(options, "optimize.spacegroup") ? cells.size() : 1;
  auto total_copies = stoi(options.at("copies"));
  auto ncopies = istrue(options, "optimize.copies") ? total_copies : 1;
  auto nmod =
      istrue(options, "optimize.model") || istrue(options, "optimize.assembly")
          ? models.size()
          : 1;
  auto nattempt = stoi(options.at("attempts"));
  // evolutionary optimization
  auto interpolation_rate =
      stod(options.at("evolution.mutation.interpolation.rate"));
  auto translation_rate =
      stod(options.at("evolution.mutation.translation.rate"));
  auto translation_size =
      stod(options.at("evolution.mutation.translation.size"));
  auto rotation_rate = stod(options.at("evolution.mutation.rotation.rate"));
  auto rotation_size = stod(options.at("evolution.mutation.rotation.size"));
  auto crossover_rate = stod(options.at("evolution.mutation.crossover.rate"));
  auto modelswap_rate = stod(options.at("evolution.mutation.modelswap.rate"));
  auto targetswap_rate = stod(options.at("evolution.mutation.cellswap.rate"));
  auto babinet_rate = stod(options.at("evolution.mutation.babinet.rate"));
  auto ksol_mut_size = stod(options.at("evolution.mutation.babinet.k.size"));
  auto ksol_min = stod(options.at("evolution.mutation.babinet.k.min"));
  auto ksol_max = stod(options.at("evolution.mutation.babinet.k.max"));
  auto bsol_mut_size = stod(options.at("evolution.mutation.babinet.B.size"));
  auto bsol_min = stod(options.at("evolution.mutation.babinet.B.min"));
  auto bsol_max = stod(options.at("evolution.mutation.babinet.B.max"));

  auto uniform = std::uniform_real_distribution<>();
  auto gaussian = std::normal_distribution<>();

  auto fixed_axes = std::array<bool, 3>{
      {options.at("evolution.mutation.translation.range.x") == "0.0",
       options.at("evolution.mutation.translation.range.y") == "0.0",
       options.at("evolution.mutation.translation.range.z") == "0.0"}};

  auto popsize = stoi(options.at("evolution.population"));
  auto sign = maximizing ? -1.0 : 1.0;
  auto start_surv_frac = stod(options.at("evolution.selection.rate.starting"));
  auto end_surv_frac = stod(options.at("evolution.selection.rate.ending"));
  auto maxgen = stoi(options.at("evolution.generations"));
  auto trans_scale = stod(options.at("transform.scale"));
  auto fft_sample = stod(options.at("transform.fft.sampling"));

  for (auto icel = size_t{0}; icel < cells.size(); icel += ncel) {
    best_solutions.emplace_back();
    best_scores.emplace_back();
    best_rfactors.emplace_back();
    structure.resize(initial_structure_size);
    partial_structure_centers.resize(initial_structure_size);
    auto frac_centers =
        vector<std::array<double, 3>>(); // model centers, fractional
    for (auto& c : ortho_centers) {
      frac_centers.emplace_back(cells[icel].fractionalize(c));
    }
    auto randomize_model = istrue(options, "optimize.model");
    auto optimize_orientation = istrue(options, "optimize.orientation");
    auto optimize_position = istrue(options, "optimize.position");
    auto model_max = models.size(); // models vector may later increase in size
                                    // if there are segments
    for (auto imod = size_t{0}; imod < model_max; imod += nmod) {
      best_solutions.back().emplace_back();
      best_scores.back().emplace_back();
      best_rfactors.back().emplace_back();
      if (options.at("assembly") == "false") {
        structure.resize(initial_structure_size);
        partial_structure_centers.resize(initial_structure_size);
      }
      if (structure.size() == 0) {
        fix_polar_axes(fixed_axes, cells[icel].space_group());
        if (ncopies == 1 && fixed_axes[0] && fixed_axes[1] && fixed_axes[2]) {
          if (user_options.count("evolution.mutation.rotation.notrans.rate") ==
              0) {
            rotation_rate =
                stod(options.at("evolution.mutation.rotation.notrans.rate"));
          }
          if (user_options.count("evolution.mutation.rotation.notrans.size") ==
              0) {
            rotation_size =
                stod(options.at("evolution.mutation.rotation.notrans.size"));
          }
        }
      }
      if (report_steps) {
        outs << "Search model initialization " << flush;
      }
      calc_transforms(models, imod, nmod, high_res_limit, trans_scale,
                      fft_sample, transforms);
      if (report_steps) {
        outs << "completed" << endl;
      }
      auto selected_target_id = size_t{0};
      for (auto icopy = 0; icopy < total_copies; icopy += ncopies) {
        // fix the cell after first copy found (only has an effect if ncel > 1)
        if (icopy > 0) {
          icel = selected_target_id;
        }
        if (structure.size() > 0) {
          if (report_steps) {
            outs << "Calculating static structure contribution" << endl;
          }
          calc_static(structure, cells[icel], hkl_permutations[icel], refl,
                      high_res_limit, fft_sample);
        }
        auto nseg =
            istrue(options, "optimize.assembly") ? nmod * ncopies : ncopies;
        auto best_solution = solution{};
        auto best_score = maximizing ? numeric_limits<double>::lowest()
                                     : numeric_limits<double>::max();
        auto best_rfactor = numeric_limits<double>::max();
        auto all_scores = vector<double>();
        for (auto iattempt = 0; iattempt < nattempt; iattempt++) {
          if (report_steps) {
            if (cells.size() > 1) {
              // cell is fixed after first copy during sequential multi-copy
              // search
              if (ncel > 1 && icopy == 0) {
                outs << "Unit cells " << (icel + 1) << '-' << (icel + ncel);
              } else {
                outs << "Unit Cell " << icel + 1 << " ("
                     << cells[icel].space_group().symbol() << ")";
              }
              outs << ", ";
            }
            if (model_max > 1) {
              if (nmod > 1) {
                outs << "Models " << (imod + 1) << '-' << (imod + nmod);
              } else {
                outs << "Model " << (imod + 1);
              }
              outs << ", ";
            }
            if (total_copies > 1) {
              if (ncopies > 1) {
                outs << "Copies " << (icopy + 1) << '-' << (icopy + ncopies);
              } else {
                outs << "Copy " << (icopy + 1);
              }
              outs << ", ";
            }
            outs << "Attempt " << (iattempt + 1) << endl;
          }
          auto sol = solution{};
          if (options.at("optimizer.global") == "evolutionary") {
            optimization::evolutionary_optimizer<solution> evolve;
            if (istrue(options, "target.global.interpolate-fc")) {
              evolve.score = scorer<pearson<float>, interpolation::Linear,
                                    global_reflection_parallelism>(
                  refl, transforms, cells, hkl_permutations);
            } else {
              evolve.score = scorer<pearson<float>, interpolation::None,
                                    global_reflection_parallelism>(
                  refl, transforms, cells, hkl_permutations);
            }
            evolve.create = [&](solution& s) {
              s.segments.resize(nseg);
              for (auto i = size_t{0}; i < nseg; ++i) {
                auto& seg = s.segments[i];
                if (randomize_model)
                  seg.model_id = imod + (size_t)(uniform(rng) * nmod);
                else
                  seg.model_id = imod + (i % nmod);
                seg.base_model_id = seg.model_id;
                if (optimize_orientation) {
                  auto quat = geom::make_quat(uniform, rng);
                  seg.orientation = quat;
                }
                if (optimize_position) {
                  seg.position = {{uniform(rng), uniform(rng), uniform(rng)}};
                } else {
                  seg.position = frac_centers[i];
                }
              }
              if (fixed_axes[0])
                s.segments[0].position[0] = frac_centers[0][0];
              if (fixed_axes[1])
                s.segments[0].position[1] = frac_centers[0][1];
              if (fixed_axes[2])
                s.segments[0].position[2] = frac_centers[0][2];
              s.target_id = icel;
              // sample the range of cells only if this is the first copy of the
              // molecule
              if (icopy == 0 && ncel > 1) {
                s.target_id += static_cast<size_t>(uniform(rng) * ncel);
              }
              s.ksol = 0.8f;   // float(uniform(rng) * 1.5);
              s.bsol = 200.0f; // float(uniform(rng) * 300.0);
            };
            evolve.procreate = [&](size_t member_index, vector<solution>& pop,
                                   const vector<size_t>& indices,
                                   const size_t nparents) {
              auto& s = pop[member_index];
              const auto& p1 = pop[indices[static_cast<size_t>(
                  floor(nparents * uniform(rng)))]];
              const auto& p2 = pop[indices[static_cast<size_t>(
                  floor(nparents * uniform(rng)))]];
              s = p1;
              for (auto i = size_t{0}; i < s.segments.size(); ++i) {
                auto& seg = s.segments[i];
                if (uniform(rng) < interpolation_rate) {
                  geom::slerp(seg.orientation, p2.segments[i].orientation,
                              uniform(rng), seg.orientation);
                  geom::interpolate(seg.position, p2.segments[i].position,
                                    uniform(rng), seg.position);
                } else if (uniform(rng) < crossover_rate) {
                  if (uniform(rng) < 0.5)
                    seg.orientation = p2.segments[i].orientation;
                  else
                    seg.position = p2.segments[i].position;
                }
                if (uniform(rng) < rotation_rate) {
                  auto alpha =
                      fmod(std::abs(gaussian(rng) * rotation_size), 1.0);
                  auto random_orientation = geom::make_quat(uniform, rng);
                  geom::slerp(seg.orientation, random_orientation, alpha,
                              seg.orientation);
                }
                for (auto j = 0; j < 3; ++j) {
                  if ((i > 0 || fixed_axes[j] == false) &&
                      uniform(rng) < translation_rate) {
                    seg.position[j] += gaussian(rng) * translation_size;
                  }
                }
                canonicalize(seg.position);
                if (uniform(rng) < modelswap_rate)
                  seg.base_model_id = seg.model_id =
                      imod + static_cast<size_t>(uniform(rng) * nmod);
              }
              if (uniform(rng) < targetswap_rate)
                s.target_id = icel + static_cast<size_t>(uniform(rng) * ncel);
              if (uniform(rng) < babinet_rate) {
                // reflects at kmin, wraps at kmax
                s.ksol = static_cast<float>(
                    ksol_min +
                    std::abs(fmod(s.ksol - ksol_min +
                                      (gaussian(rng) * ksol_mut_size),
                                  ksol_max - ksol_min)));
              }
              if (uniform(rng) < babinet_rate) {
                s.bsol = static_cast<float>(
                    bsol_min +
                    std::abs(fmod(s.bsol - bsol_min +
                                      (gaussian(rng) * bsol_mut_size),
                                  bsol_max - bsol_min)));
              }
            };
            if (options.at("report.evolution") == "progress") {
              evolve.report = [&](int gen, const vector<solution>& sol,
                                  const vector<double>& scores) {
                if (gen == 1)
                  outs << "Global search ";
                outs << '.';
                if (gen == maxgen)
                  outs << " completed\n";
                outs << flush;
              };
            } else if (options.at("report.evolution") == "generations") {
              evolve.report = [&](int gen, const vector<solution>& sol,
                                  const vector<double>& scores) {
                outs << fixed << setprecision(3) << showpoint;
                outs << "Generation " << setw(2) << gen;
                outs << ": Best score = "
                     << (maximizing
                             ? *max_element(scores.cbegin(), scores.cend())
                             : *min_element(scores.cbegin(), scores.cend()));
                outs << ", Ave. score = " << calc_mean(scores.cbegin(),
                                                       scores.cend());
                outs << ", Std. dev. = " << sqrt(calc_population_variance(
                                                scores.cbegin(), scores.cend()))
                     << endl;
              };
            }
            evolve.select = [=](int gen) {
              return start_surv_frac +
                     ((end_surv_frac - start_surv_frac) * gen / maxgen);
            };
            evolve.converged = [=](int gen, const vector<double>& scores) {
              return gen == maxgen;
            };
            evolve.rank = [=](size_t index, const vector<double>& scores) {
              return scores[index] * sign;
            };
            eo_platform_config<global_population_parallelism>::configure(
                evolve);

            sol = evolve(popsize);
          } else {
            sol.target_id = icel;
            sol.ksol = 0.8f;
            sol.bsol = 200.0f;
            sol.segments.resize(nseg);
            // to work with "optimize.assembly" and "optimize.model" options
            auto iseg = imod;
            for (auto& seg : sol.segments) {
              seg.base_model_id = seg.model_id = iseg++;
              seg.position = frac_centers[seg.model_id];
            }
          }
          auto interp_cc = scorer<pearson<float>, interpolation::Linear,
                                  local_reflection_parallelism>(
              refl, transforms, cells, hkl_permutations);
          auto rfactor = scorer<scaled_rfactor<float>, interpolation::Linear,
                                local_reflection_parallelism>(
              refl, transforms, cells, hkl_permutations);
          if (options.at("optimizer.local") == "simplex") {
            if (istrue(options, "report.local")) {
              outs << "Local optimization" << flush;
            }
            dsimplex_optimize(sol, interp_cc, stod(options.at("simplex.size")),
                              stod(options.at("simplex.tolerance")),
                              optimize_orientation, optimize_position, true);
            if (istrue(options, "report.local")) {
              outs << " completed" << endl;
            }
            if (istrue(options, "optimize.local-segments")) {
              if (split_segments(sol, models, cells, frac_centers)) {
                if (istrue(options, "report.local")) {
                  outs << "Local optimization of segments" << flush;
                }
                calc_segment_transforms(sol, transforms, models, high_res_limit,
                                        trans_scale, fft_sample);
                dsimplex_optimize(
                    sol, interp_cc, stod(options.at("simplex.size")),
                    stod(options.at("simplex.tolerance")), optimize_orientation,
                    optimize_position, true);
                if (istrue(options, "report.local")) {
                  outs << " completed" << endl;
                }
              }
            }
          }
          if (istrue(options, "report.solution.normalized")) {
            auto is_local_search =
                options.at("optimizer.global") == "none" ||
                options.at("optimize.orientation") == "false" ||
                options.at("optimize.position") == "false";
            normalize_solution(sol, cells, frac_centers, fixed_axes,
                               structure.size() > 0, is_local_search);
          }
          auto current_score = interp_cc(sol);
          auto current_rfactor = rfactor(sol);

          selected_target_id = sol.target_id;

          all_scores.emplace_back(current_score);
          if ((maximizing && current_score > best_score) ||
              ((!maximizing) && current_score < best_score)) {
            best_solution = sol;
            best_score = current_score;
            best_rfactor = current_rfactor;
          }

          // temporarily add this solution to the stored structure
          // for clash testing and coordinate output
          auto strsize = structure.size();
          append_structure(sol, structure, models, cells,
                           partial_structure_centers);

          auto clashes = -1;
          if (istrue(options, "report.solution")) {
            if (options.at("packing.clash.type") == "interatomic") {
              clashes = calc_packing_clashes(
                  sol, models, structure, cells,
                  options.at("packing.clash.atoms") == "all",
                  stod(options.at("packing.clash.distance")));
            }
            report_solution(sol, outs, current_score, current_rfactor, iattempt,
                            cells, clashes);
          }
          if (options.at("report.stats") == "running" ||
              (iattempt == (nattempt - 1) &&
               options.at("report.stats") == "final")) {
            report_stats(imod, icel, icopy, iattempt, all_scores, outs);
          }
          if (options.at("output.level") == "all") {
            auto path = string{};
            if (options.at("output.prefix") != "-") {
              path = options.at("output.prefix") + '.';
              if (ncel < cells.size())
                path += to_string(icel + 1) + '.';
              if (nmod < model_max)
                path += to_string(imod + 1) + '.';
              if (ncopies < total_copies)
                path += to_string(icopy + 1) + '.';
              path += to_string(iattempt + 1);
              path += ".pdb";
            }
            if (report_steps && path.length() > 0) {
              outs << "Writing coordinates to " << path << '\n' << endl;
            }
            write_solution(structure, cells[sol.target_id], current_score,
                           current_rfactor, clashes, path, outs,
                           istrue(options, "output.reassign-chain-ids"));
          }
          structure.resize(strsize);
          partial_structure_centers.resize(strsize);
        }

        best_solutions.back().back().emplace_back(best_solution);
        best_scores.back().back().emplace_back(best_score);
        best_rfactors.back().back().emplace_back(best_rfactor);

        fixed_axes = {
            {options.at("evolution.mutation.translation.range.x") == "0.0",
             options.at("evolution.mutation.translation.range.y") == "0.0",
             options.at("evolution.mutation.translation.range.z") == "0.0"}};
        rotation_rate = stod(options.at("evolution.mutation.rotation.rate"));
        rotation_size = stod(options.at("evolution.mutation.rotation.size"));

        append_structure(best_solution, structure, models, cells,
                         partial_structure_centers);
      }
    }
  }
  if (options.at("output.level") == "best") {
    auto best_score = maximizing ? numeric_limits<double>::lowest()
                                 : numeric_limits<double>::max();
    auto best_cell_index = size_t{0};
    auto best_model_index = size_t{0};
    for (auto i = size_t{0}; i < best_scores.size();
         ++i) { // for each cell choice
      auto mstart =
          (istrue(options, "assembly")) ? best_scores[i].size() - 1 : 0;
      for (auto j = mstart; j < best_scores[i].size();
           ++j) { // for each model or final model if assembly
        auto score = best_scores[i][j].back(); // get best score for final copy
        if ((maximizing && score > best_score) ||
            ((!maximizing) && score < best_score)) {
          best_cell_index = i;
          best_model_index = j;
          best_score = score;
        }
      }
    }
    auto best_rfactor = best_rfactors[best_cell_index][best_model_index].back();
    auto overall_best =
        best_solutions[best_cell_index][best_model_index].back();
    auto path = string{};
    if (options.at("output.prefix") != "-") {
      path = options.at("output.prefix") + ".best.pdb";
    }
    structure.resize(initial_structure_size);
    partial_structure_centers.resize(initial_structure_size);
    auto mstart = (istrue(options, "assembly")) ? 0 : best_model_index;
    auto mlimit = (istrue(options, "assembly"))
                      ? best_solutions[best_cell_index].size()
                      : best_model_index + 1;
    for (auto i = mstart; i < mlimit;
         ++i) { // for best model or each model if assembly
      auto cplimit = best_solutions[best_cell_index][i].size();
      for (auto j = size_t{0}; j < cplimit; ++j) { // for each copy
        append_structure(best_solutions[best_cell_index][i][j], structure,
                         models, cells, partial_structure_centers);
      }
    }
    // have to recalculate clashes
    auto clashes = -1;
    if (options.at("packing.clash.type") == "interatomic") {
      clashes =
          calc_packing_clashes(overall_best, models, structure, cells,
                               options.at("packing.clash.atoms") == "all",
                               stod(options.at("packing.clash.distance")));
    }
    if (report_steps && path.length() > 0) {
      outs << "Writing coordinates to " << path << '\n' << endl;
    }
    write_solution(structure, cells[overall_best.target_id], best_score,
                   best_rfactor, clashes, path, outs,
                   istrue(options, "output.reassign-chain-ids"));
  }
  if (istrue(options, "report.time.total")) {
    auto epmr_end_time = std::chrono::steady_clock::now();
    outs << "Total elapsed time: "
         << std::chrono::duration<double>(epmr_end_time - epmr_start_time)
                .count() << " sec" << endl;
  }
}

// check for special conditions and consistency of options
void reconcile_options(unordered_map<string, string>& options,
                       const unordered_map<string, string>& user_options) {
  if ((options.at("optimizer.global") == "evolutionary") &&
      (stoi(options.at("evolution.generations")) == 0)) {
    options.at("optimizer.global") = "none";
  }
  // change attempts to one if no global optimizer
  if (options.at("optimizer.global") == "none") {
    if (user_options.count("attempts") == 0)
      options.at("attempts") = "1";
  }
  if (istrue(options, "optimize.spacegroup")) {
    if (options.at("optimizer.global") == "none") {
      throw runtime_error(
          "space group search (-C) and local search (-g0) cannot be combined.");
    }
    if (options.at("structure.source") != "") {
      throw runtime_error("partial structure (-s) is incompatible with space "
                          "group search (-C).");
    }
    // increase population size and number of generations if searching multiple
    // spacegroups
    if (user_options.count("evolution.generations") == 0)
      options.at("evolution.generations") = "80";
    if (user_options.count("evolution.population") == 0)
      options.at("evolution.population") = "600";
  }
  if (istrue(options, "optimize.assembly") ||
      istrue(options, "optimize.copies")) {
    if (options.at("optimizer.global") == "none") {
      throw runtime_error("multi-copy search (-M or -A) is incompatible with "
                          "local search (-g0).");
    }
    // increase the number of generations if searching multiple
    // molecules/fragments
    if (user_options.count("evolution.generations") == 0)
      options.at("evolution.generations") = "400";
  }
  if (istrue(options, "optimize.model")) {
    if (options.at("optimizer.global") == "none") {
      throw runtime_error("model selection during search (-S) is incompatible "
                          "with local search (-g0).");
    }
  }
  if (user_options.count("output.reassign-chain-ids") == 0 &&
      (!istrue(options, "assembly") && stoi(options.at("copies")) == 1)) {
    options.at("output.reassign-chain-ids") = "false";
  }
  if (options.at("target.type") == "ccf") {
    options.at("optimizer.goal") = "maximum";
  } else if (options.at("target.type") == "rfactor") {
    options.at("optimizer.goal") = "minimum";
  }

  if (options.at("optimize.orientation") != "true") {
    options.at("evolution.mutation.rotation.type") = "none";
    options.at("evolution.mutation.rotation.rate") = "0.0";
  }
  if (options.at("optimize.position") != "true") {
    options.at("evolution.mutation.translation.type") = "none";
    options.at("evolution.mutation.translation.rate") = "0.0";
    options.at("evolution.mutation.translation.range.x") = "0.0";
    options.at("evolution.mutation.translation.range.y") = "0.0";
    options.at("evolution.mutation.translation.range.z") = "0.0";
  }
}

void print_epmr_header(ostream& output_stream) {
  output_stream << "EPMR Version " << epmr_defaults.at("version") << ", "
                << __DATE__ << "\nThe latest version of this program is always "
                               "available from www.epmr.org.\n" << endl;
}

void print_option_summary(unordered_map<string, string>& options,
                          ostream& outs) {
  outs << "Optimizing:";
  if (istrue(options, "optimize.orientation"))
    outs << " orientation";
  if (istrue(options, "optimize.position"))
    outs << " position";
  if (istrue(options, "optimize.model"))
    outs << " model";
  if (istrue(options, "optimize.assembly"))
    outs << " assembly";
  if (istrue(options, "optimize.spacegroup"))
    outs << " space-group";
  if (istrue(options, "optimize.copies"))
    outs << " multiple-copies";
  outs << '\n';
  if (istrue(options, "assembly"))
    outs << "The set of search models will be treated as an assembly.\n";
  outs << "Resolution range: " << options.at("resolution.low") << " - "
       << options.at("resolution.high") << " Angstroms\n";
  outs << "Copies of molecule to find: " << options.at("copies") << '\n';
  outs << "Number of attempts (per copy): " << options.at("attempts") << '\n';
  /*
      outs << "Threshold score to end search: " << options["target.threshold"]
     << '\n';
      auto bump = stod(options.at("bump.threshold"));
      if (bump > 0.0)
          outs << "Minimum unpenalized center-to-center packing distance: " <<
     bump << " Angstroms\n";
  */
  if (options.at("optimizer.global") == "evolutionary") {
    outs << "Number of generations: " << options.at("evolution.generations")
         << '\n';
    outs << "Population size: " << options.at("evolution.population") << '\n';
  }
  outs << "Random number seed: " << options.at("random.seed") << '\n';
  outs << "Prefix for output coordinate file name: ";
  const auto& prefix = options.at("output.prefix");
  if (prefix == "-") {
    outs << "none (stdout)";
  } else {
    outs << prefix;
  }
  outs << '\n' << endl;
}

/*
std::array<double, 3> calc_cns_euler_angles(const std::array<double, 4>& q) {
    auto angles = std::array<double, 3>();
    auto rot = geom::make_matrix33(q);

    auto c2 = max(-1.0, min(1.0, rot[2][2]));
    angles[1] = acos(c2);

    auto s2 = sqrt(max(0.0, 1.0 - rot[2][2] * rot[2][2]));

    if (s2 > 0.0000001) {
        auto s3 = rot[0][2] / s2;
        auto c3 = rot[1][2] / s2;
        c3 = max(-1.0, min(1.0, c3));
        angles[2] = acos(c3);

        if (s3 < 0.0) {
            angles[2] = 2.0*geom::PI - angles[2];
        }

        auto s1 = rot[2][0] / s2;
        auto c1 = -rot[2][1] / s2;
        c1 = max(-1.0, min(1.0, c1));
        angles[0] = acos(c1);

        if (s1 < 0.0) {
            angles[0] = 2.0*geom::PI - angles[0];
        }
    } else {
        angles[2] = 0.0;

        auto c1 = rot[0][0];
        auto s1 = rot[1][0];
        c1 = max(-1.0, min(1.0, c1));
        angles[0] = acos(c1);

        if (s1 < 0.0) {
            angles[0] = 2.0 * geom::PI - angles[0];
        }
    }
    return angles;
}
*/

// based on code from Ken Shoemake in "Graphics Gems IV",
// Paul S. Heckbert, ed., Academic Press Professional, Toronto, 1994
std::array<double, 3>
calc_ccp4_euler_angles(const std::array<double, 4>& quat) {
  auto ea = std::array<double, 3>{};
  auto m = geom::make_matrix33(quat);
  auto sy = sqrt(m[2][1] * m[2][1] + m[2][0] * m[2][0]);
  if (sy > 16 * numeric_limits<double>::epsilon()) {
    ea = {{atan2(m[2][1], m[2][0]), atan2(sy, m[2][2]),
           atan2(m[1][2], -m[0][2])}};
  } else {
    ea = {{atan2(-m[1][0], m[1][1]), atan2(sy, m[2][2]), 0}};
  }
  ea[0] = -ea[0];
  ea[1] = -ea[1];
  ea[2] = -ea[2];
  if (ea[1] < 0.0) {
    ea[0] = geom::PI + ea[0];
    ea[1] = -ea[1];
    ea[2] = geom::PI + ea[2];
  }
  if (ea[0] < 0.0)
    ea[0] += geom::PI * 2;
  if (ea[2] < 0.0)
    ea[2] += geom::PI * 2;
  auto t = ea[0];
  ea[0] = ea[2];
  ea[2] = t;
  return ea;
}

void report_solution(const solution& s, ostream& outs, double score,
                     double rfactor, int attempt, vector<unit_cell>& cells,
                     int clashes) {
  outs << fixed << showpoint << "\n          #    CC    R    d<     al     be  "
                                "   ga     x      y      z "
                                "  mod sg\n";
  auto i = 0;
  for (const auto& seg : s.segments) {
    outs << "Solution " << setw(2) << (attempt + 1);
    if (s.segments.size() > 1)
      outs << (char)('a' + i % 26);
    else
      outs << ' ';
    outs << ' ';
    if (i == 0) {
      outs << setprecision(3) << score << ' ' << rfactor << ' ' << setw(3)
           << clashes << ' ' << " ";
    } else {
      outs << "                 ";
    }
    auto euler = calc_ccp4_euler_angles(seg.orientation);
    outs << setprecision(1) << setw(6) << euler[0] / geom::PI * 180.0 << ' '
         << setw(6) << euler[1] / geom::PI * 180.0 << ' ' << setw(6)
         << euler[2] / geom::PI * 180.0 << ' ' << setprecision(3) << setw(6)
         << seg.position[0] << ' ' << setw(6) << seg.position[1] << ' '
         << setw(6) << seg.position[2] << ' ' << setw(2)
         << seg.base_model_id + 1 << ' ';
    const auto sg = cells[s.target_id].space_group().number();
    //        const auto space = 7 - (int) sg.length();
    //        for (auto j = 0; j < space / 2 + (space % 2); ++j) {
    //            outs << ' ';
    //        }
    outs << setw(3) << sg;
    //        for (auto j = 0; j < space / 2; ++j) {
    //            outs << ' ';
    //        }
    outs /*<< ' ' << score << ' ' << rfactor << ' ' << setw(2) << clashes*/
        << '\n';
    //        euler = calc_cns_euler_angles(seg.orientation);
    //        outs << "   CNS Euler angles: " << setprecision(2) <<
    //            euler[0] / geom::PI * 180.0 << ' ' <<
    //            euler[1] / geom::PI * 180.0 << ' ' <<
    //            euler[2] / geom::PI * 180.0 << endl;
    ++i;
  }
  outs << endl;
  //    if (clashes >= 0) outs << "   Close CA-CA contacts: " << clashes <<
  //    '\n';
  //    outs << "   Solvent correction parameters: Ksol = " << s.ksol << ", Bsol
  //    = " << s.bsol << '\n' << endl;
}

void write_solution(const vector<vector<atom>>& structure,
                    const unit_cell& cell, double score, double rfactor,
                    int clashes, const string& path, ostream& outstr,
                    bool reassign_chains) {
  if (path.length() > 0) {
    std::ofstream filestream(path, ios_base::out);
    write_atoms(filestream, score, rfactor, clashes, cell, structure,
                reassign_chains);
  } else {
    write_atoms(outstr, score, rfactor, clashes, cell, structure,
                reassign_chains);
  }
}

void write_atoms(ostream& coordstream, double score, double rfactor,
                 int packing_clashes, const unit_cell& cell,
                 const vector<vector<atom>>& structure, bool reassign_chains) {

  coordstream << fixed << showpoint << setprecision(3);
  coordstream << "TITLE     EPMR Score: " << score << " Rfactor: " << rfactor;
  coordstream << " Packing clashes: " << packing_clashes << "     \n";

  coordstream << "CRYST1" << setprecision(3) << setw(9) << cell.a_axis()
              << setw(9) << cell.b_axis() << setw(9) << cell.c_axis()
              << setprecision(2) << setw(7) << cell.alpha() * 180.0 / geom::PI
              << setw(7) << cell.beta() * 180.0 / geom::PI << setw(7)
              << cell.gamma() * 180.0 / geom::PI << ' ' << setw(11) << std::left
              << cell.space_group().hm_symbols()[0] << std::right << setw(4)
              << cell.space_group().symmetry_operators().size() << '\n';

  auto chain_index = 0;
  auto f2ct = geom::make_matrix44(cell.orthogonalization_matrix());
  for (const auto& str : structure) {
    for (const auto& atm : str) {
      auto chainid =
          reassign_chains ? (char)('A' + (chain_index % 26)) : atm.line[21];
      write_atom_record(atm, f2ct, chainid, coordstream);
    }
    // coordstream << "TER                                                   "
    // << endl;
    chain_index++;
  }
  coordstream << "END                                                   "
              << endl;
}

void write_atom_record(const atom& a,
                       const std::array<std::array<double, 4>, 4>& trnsfrm,
                       char chainid, ostream& coordstream) {
  coordstream << a.line.substr(0, 21);
  coordstream << chainid;
  coordstream << a.line.substr(22, 8);
  auto xyz = geom::transform(trnsfrm, a.position);
  coordstream << setprecision(3);
  coordstream << setw(8) << xyz[0];
  coordstream << setw(8) << xyz[1];
  coordstream << setw(8) << xyz[2];
  if (a.line.length() > 54) {
    coordstream << a.line.substr(54) << '\n';
  } else {
    coordstream << "                          " << '\n';
  }
}

// dispose of any previously used transforms and create new ones
void calc_transforms(
    const vector<vector<atom>>& models, size_t firstindex, size_t n,
    double highres_limit, double scale, double sampling,
    vector<unique_ptr<molecular_transform<float>>>& transforms) {
  transforms.resize(models.size());
  for (auto i = size_t{0}; i < firstindex; ++i) {
    transforms[i] = nullptr;
  }
  auto limit = firstindex + n;
  for (auto i = firstindex; i < limit; ++i) {
    if (transforms[i] == nullptr) {
      transforms[i] =
          unique_ptr<molecular_transform<float>>(new molecular_transform<float>(
              models[i], highres_limit, scale, sampling));
    }
  }
}

void calc_segment_transforms(
    const solution& sol,
    vector<unique_ptr<molecular_transform<float>>>& transforms,
    const vector<vector<atom>>& models, double highres, double scale,
    double sampling) {
  for (const auto& seg : sol.segments) {
    auto model_id = seg.model_id;
    if (transforms.size() <= model_id)
      transforms.resize(model_id + 1);
    if (transforms[model_id] == nullptr) {
      transforms[model_id] =
          unique_ptr<molecular_transform<float>>(new molecular_transform<float>(
              models[model_id], highres, scale, sampling));
    }
  }
}

void append_structure(const solution& sol, vector<vector<atom>>& structure,
                      const vector<vector<atom>>& models,
                      const vector<unit_cell>& cells,
                      vector<std::array<double, 3>>& centers) {
  for (const auto& seg : sol.segments) {
    auto ortho_trans = cells[sol.target_id].orthogonalize(seg.position);
    auto transform = geom::make_matrix44(seg.orientation, ortho_trans, 1.0);
    structure.emplace_back(models[seg.model_id]);
    auto coord_ave = std::array<double, 3>{};
    for (auto& a : structure.back()) {
      std::array<double, 3> txyz = geom::transform(transform, a.position);
      a.position = cells[sol.target_id].fractionalize(txyz);
      coord_ave = geom::add(coord_ave, a.position);
    }
    geom::scale(coord_ave, 1.0 / structure.back().size(), coord_ave);
    centers.push_back(coord_ave);
  }
}

vector<unit_cell> load_cells(const string& path) {
  auto format = string("text");
  if (path.rfind(".mtz") == path.size() - 4) {
    format = "mtz";
  }
  if (path.rfind(".sca") == path.size() - 4) {
    format = "scalepack";
  }
  auto cell = std::array<double, 6>{};
  auto sgnum = 0;
  // MTZ format
  if (format == "mtz") {
    ifstream mtz(path, ios_base::in | ios_base::binary);
    if (mtz.fail()) {
      const auto errmsg = "unable to open the file ";
      throw ios_base::failure(errmsg + path);
    }
    mtz.exceptions(ios_base::badbit);
    if (!mtz_is_valid(mtz)) {
      throw runtime_error("the file is not a valid MTZ file");
    }
    mtz_cell_constants(mtz, cell);
    sgnum = mtz_sgnum(mtz);
  } else if (format == "scalepack") {
    ifstream cellstream(path);
    if (cellstream.fail()) {
      const auto errmsg = "unable to open the file ";
      throw ios_base::failure(errmsg + path);
    }
    auto spcgrp_str = string{};
    cellstream.exceptions(ios_base::badbit);
    cellstream.ignore(1000, '\n');
    cellstream.ignore(1000, '\n');
    cellstream >> cell[0] >> cell[1] >> cell[2] >> cell[3] >> cell[4] >>
        cell[5] >> spcgrp_str;
    sgnum = spacegroup::symbol_to_number(spcgrp_str);
  }
  // free-text format
  else if (format == "text") {
    ifstream cellstream(path);
    if (cellstream.fail()) {
      const auto errmsg = "unable to open the file ";
      throw ios_base::failure(errmsg + path);
    }
    cellstream.exceptions(ios_base::badbit);
    cellstream >> cell[0] >> cell[1] >> cell[2] >> cell[3] >> cell[4] >>
        cell[5] >> sgnum;
  }
  for (auto p : cell) {
    if (p <= 0.0) {
      throw runtime_error("bad cell definition in input file");
    }
  }
  if (sgnum < 1 || sgnum > 230) {
    throw runtime_error("bad or missing space group information in input file");
  }
  return vector<unit_cell>{
      unit_cell(sgnum, cell[0], cell[1], cell[2], cell[3], cell[4], cell[5])};
}

void load_models(const string& path, vector<vector<atom>>& models) {
  if (path[0] == '@') {
    ifstream name_stream(path.substr(1));
    if (name_stream.fail())
      throw ios_base::failure("unable to open the file " + path.substr(1));
    name_stream.exceptions(ios_base::badbit);
    auto nextpath = string{};
    // process listed files, which also can be lists if prepended with '@'
    while (name_stream >> nextpath) {
      load_models(nextpath, models);
    }
  } else if (path == "-") {
    load_model(std::cin, models, "standard input");
  } else {
    ifstream atomstream(path);
    load_model(atomstream, models, path);
  }
}

void load_model(istream& atomstream, vector<vector<atom>>& models,
                const string& path) {
  if (atomstream.fail()) {
    const auto errmsg = string{"unable to open the file "};
    throw ios_base::failure(errmsg + path);
  }
  atomstream.exceptions(ios_base::badbit);
  models.emplace_back(std::istream_iterator<atom>(atomstream),
                      std::istream_iterator<atom>());
}

void calc_static(const vector<vector<atom>>& structure, const unit_cell& cell,
                 unsigned int hkl_permutation, vector<reflection>& ref,
                 double res_limit, double sampling) {
  if (structure.size() > 0) {
    auto map = fourier_map<float>(cell, numeric_limits<double>::max(),
                                  res_limit, sampling);
    for (const auto& str : structure) {
      for (const auto& atom : str) {
        map.add_atom(atom.position[0], atom.position[1], atom.position[2],
                     atom.occ, atom.b, atom.Z);
      }
    }
    map.invert();
    auto h_ind = hkl_permutation;
    auto k_ind = (hkl_permutation + 1) % 3;
    auto l_ind = (hkl_permutation + 2) % 3;
    for (auto& r : ref) {
      r.fpartial = map.get_structure_factor(
          int(r.hkl[h_ind]), int(r.hkl[k_ind]), int(r.hkl[l_ind]));
    }
  } else {
    for (auto& r : ref) {
      r.fpartial = complex<float>{0.0f, 0.0f};
    }
  }
}

void report_stats(size_t model, size_t spcgrp, int copy, int attempt,
                  vector<double>& scores, ostream& ostr) {
  static auto header_printed = false;
  if (!header_printed) {
    ostr << "SpcGrp\tModel\tCopy\tAttempt\tScore\tMin\tMean\tMax\tNH-Mean\tRate"
         << '\n';
    header_printed = true;
  }
  auto ave = mean<double>{};
  auto nonmaxmean = mean<double>{};
  auto minim = minimum<double>{};
  auto maxim = maximum<double>{};
  for (auto score : scores) {
    ave.add(score);
    minim.add(score);
    maxim.add(score);
  }
  auto maxcnt = size_t{0};
  auto maxval = maxim();
  for (auto score : scores) {
    if (maxval - score < 0.002) {
      maxcnt++;
    } else {
      nonmaxmean.add(score);
    }
  }
  ostr << fixed << setprecision(3) << showpoint;
  ostr << spcgrp << '\t' << model << '\t' << copy << '\t' << scores.size()
       << '\t' << scores.back() << '\t';
  ostr << minim() << '\t' << ave() << '\t' << maxim() << '\t'
       << (maxcnt < scores.size() ? nonmaxmean() : 0.0) << '\t'
       << ((double)maxcnt / (double)scores.size()) << endl;
}

// This function fixes axes as appropriate for the space group.
// Only the first space group is checked when ncell > 1.
// The restrictions should not applied if there is partial structure.
// (When there are multiple segments, only the first segment is affected by
// these translation options.)
void fix_polar_axes(std::array<bool, 3>& fixed_axes, const spacegroup& sg) {
  if (sg.is_a_axis_polar()) {
    fixed_axes[0] = true;
  }
  if (sg.is_b_axis_polar()) {
    fixed_axes[1] = true;
  }
  if (sg.is_c_axis_polar()) {
    fixed_axes[2] = true;
  }
}

bool split_segments(solution& sol, vector<vector<atom>>& modelvec,
                    vector<unit_cell>& cells,
                    vector<std::array<double, 3>>& centers) {
  auto original_seg_count = sol.segments.size();
  auto new_seg_list = vector<rigid_body>{};
  for (auto& seg : sol.segments) {
    // check for multiple segments and make the models
    // split based on PDB seg-id field
    auto& atoms = modelvec[seg.model_id];
    auto segmap = unordered_map<string, unique_ptr<vector<atom>>>{};
    for (const auto& a : atoms) {
      auto segid = string{};
      auto fieldwidth = (int)a.line.size() - 72;
      if (fieldwidth > 4)
        fieldwidth = 4;
      if (fieldwidth > 0)
        segid = a.line.substr(72, fieldwidth);
      else
        segid = " ";
      if (segmap.count(segid) == 0)
        segmap[segid] = unique_ptr<vector<atom>>(new vector<atom>);
      segmap[segid]->push_back(a);
    }
    // if there is more than one seg-id, subdivide the current segment
    if (segmap.size() > 1) {
      for (auto& entry : segmap) {
        modelvec.push_back(move(*entry.second));
        auto newseg = seg;
        newseg.model_id = (int)modelvec.size() - 1;
        new_seg_list.push_back(newseg);
        // The segment is given the same center as its parent segment,
        // which will be the origin of the molecular transform, since
        // the segments aren't individually recentered.
        centers.push_back(centers[seg.model_id]);
      }
    } else {
      new_seg_list.push_back(seg);
    }
  }
  sol.segments = new_seg_list;
  return (sol.segments.size() > original_seg_count);
}

// uses the method described by Sheldrick,
// http://shelx.uni-ac.gwdg.de/SHELX/cryst-alg.pdf
int calc_packing_clashes(const solution& sol,
                         const vector<vector<atom>>& models,
                         const vector<vector<atom>>& structure,
                         const vector<unit_cell>& cells, bool all_atoms,
                         double clashdist) {
  static const auto ShiftVal = 99.5;
  auto cutoff = clashdist * clashdist;
  const auto& cell = cells[sol.target_id];
  const auto& sg = cell.space_group();
  const auto nsymm = sg.num_symm_ops();
  auto clashes = 0;
  auto positions = vector<std::array<double, 3>>{};
  auto calphas = vector<const atom*>{};
  // copy all coordinates along with symmetry-related positions
  for (const auto& str : structure) {
    for (const auto& a : str) {
      if (all_atoms || ((a.line.substr(12, 4) == " CA ") && (a.Z == 6))) {
        calphas.push_back(&a);
        for (auto i = 0; i < nsymm; ++i) {
          auto pos = a.position;
          pos = geom::transform(sg.symm_op(i), pos);
          canonicalize(pos);
          positions.push_back(pos);
        }
      }
    }
  }
  const auto npos = positions.size();
  const auto shift = std::array<double, 3>{{ShiftVal, ShiftVal, ShiftVal}};
  for (auto i = size_t{0}; i < npos; i += nsymm) {
    const auto& p1 = positions[i];
    for (auto j = i + 1; j < npos; ++j) {
      auto p2 = geom::add(positions[j], shift);
      auto dx = fmod(p2[0] - p1[0], 1.0) - 0.5;
      auto dy = fmod(p2[1] - p1[1], 1.0) - 0.5;
      auto dz = fmod(p2[2] - p1[2], 1.0) - 0.5;
      auto diff = std::array<double, 3>{{dx, dy, dz}};
      diff = cell.orthogonalize(diff);
      if (geom::lengthsq(diff) < cutoff) {
        // don't look at intra-molecular interactions,
        // so check the original positions to eliminate those
        if (j % nsymm == 0) {
          auto apos = cell.orthogonalize(calphas[i / nsymm]->position);
          auto bpos = cell.orthogonalize(calphas[j / nsymm]->position);
          if (geom::distsq(apos, bpos) > cutoff) {
            ++clashes;
          }
        } else {
          ++clashes;
        }
      }
    }
  }
  return clashes;
}

void normalize_solution(solution& sol, const vector<unit_cell>& cells,
                        const vector<std::array<double, 3>>& centers,
                        const std::array<bool, 3>& fixed_axes,
                        bool have_partial_str, bool is_local_search) {
  if (is_local_search)
    normalize_local_search_solution(sol, cells, centers, fixed_axes);
  else
    normalize_global_search_solution(sol, cells, centers, fixed_axes,
                                     have_partial_str);
}

// Canonicalize solutions that underwent only a local search, or a
// rotation-only/translation-only search.
// Translations that should be fixed are reset to the starting position
// (these are allowed to move in the Downhill Simplex minimizer).
// For local searches, multiple segments will be due to seg-ids in the input
// file or the -A option.
// The position of any partial structure is ignored for local searches, but
// it is assumed that partial structure might be present, and
// the origin is not allowed to change.
void normalize_local_search_solution(
    solution& sol, const vector<unit_cell>& cells,
    const vector<std::array<double, 3>>& centers,
    const std::array<bool, 3>& fixed_axes) {
  // Reset the solution along an axis if any axes are fixed
  // because of the space group or because it was a rotation search.
  auto xdrift = 0.0;
  auto ydrift = 0.0;
  auto zdrift = 0.0;
  const auto& seg0 = sol.segments[0];
  if (fixed_axes[0]) {
    xdrift = seg0.position[0] - centers[seg0.model_id][0];
  }
  if (fixed_axes[1]) {
    ydrift = seg0.position[1] - centers[seg0.model_id][1];
  }
  if (fixed_axes[2]) {
    zdrift = seg0.position[2] - centers[seg0.model_id][2];
  }
  for (auto& seg : sol.segments) {
    geom::normalize(seg.orientation, seg.orientation);
    seg.position[0] -= xdrift;
    seg.position[1] -= ydrift;
    seg.position[2] -= zdrift;
  }
  // Large shifts out of the starting cell *could* occur during local
  // minimization
  // or from translation-only searches. To correct this, we move a segment
  // back to its original unit cell using just whole cell translations.
  // (The original position of the model could be far outside of the canonical
  // cell.)
  for (auto& seg : sol.segments) {
    auto cell_shift = std::array<double, 3>{{floor(centers[seg.model_id][0]),
                                             floor(centers[seg.model_id][1]),
                                             floor(centers[seg.model_id][2])}};
    canonicalize(seg.position);
    seg.position[0] += cell_shift[0]; // + min_trans[0];
    seg.position[1] += cell_shift[1]; // + min_trans[1];
    seg.position[2] += cell_shift[2]; // + min_trans[2];
  }
}

// Canonicalize solutions resulting from a global search.
// Translations that should be fixed are reset to the starting position
// Then symmetry operations and alternate origins are used to bring
// the solution close to the origin, but with the segment centers
// at positive x, y, z.
// There is no attempt to move the solution close to any partial
// structure that was input or to move the partial structure
// to a canonical position.
void normalize_global_search_solution(
    solution& sol, const vector<unit_cell>& cells,
    const vector<std::array<double, 3>>& centers,
    const std::array<bool, 3>& fixed_axes, bool have_partial_str) {
  auto xshift = 0.0;
  auto yshift = 0.0;
  auto zshift = 0.0;
  const auto& seg0 = sol.segments[0];
  const auto& sg = cells[sol.target_id].space_group();
  if (fixed_axes[0]) {
    if (sg.is_a_axis_polar()) {
      xshift = numeric_limits<double>::max();
      for (rigid_body& seg : sol.segments) {
        if (seg.position[0] < xshift)
          xshift = seg.position[0];
      }
    } else
      xshift = seg0.position[0] - centers[seg0.model_id][0];
  }
  if (fixed_axes[1]) {
    if (sg.is_b_axis_polar()) {
      yshift = numeric_limits<double>::max();
      for (rigid_body& seg : sol.segments) {
        if (seg.position[1] < yshift)
          yshift = seg.position[1];
      }
    } else
      yshift = seg0.position[1] - centers[seg0.model_id][1];
  }
  if (fixed_axes[2]) {
    if (sg.is_c_axis_polar()) {
      zshift = numeric_limits<double>::max();
      for (rigid_body& seg : sol.segments) {
        if (seg.position[2] < zshift)
          zshift = seg.position[2];
      }
    } else
      zshift = seg0.position[2] - centers[seg0.model_id][2];
  }
  for (rigid_body& seg : sol.segments) {
    geom::normalize(seg.orientation, seg.orientation);
    seg.position[0] -= xshift;
    seg.position[1] -= yshift;
    seg.position[2] -= zshift;
  }
  const auto zero = std::array<double, 3>{};
  // choice of origin must be the same for all segments and is fixed if there is
  // partial structure
  auto fixed_origin = std::array<double, 3>{
      {-2.0, 0.0, 0.0}}; // x value less than -1 signals no initial fixed origin
  if (have_partial_str) {
    fixed_origin[0] = 0.0;
  }
  // for multiple segments, fixed_origin is set by TransformSegmentTowardPoint
  // on
  // the first iteration through the loop and used thereafter
  for (rigid_body& seg : sol.segments) {
    transform_segment_toward_point(seg, cells[sol.target_id], zero,
                                   fixed_origin);
  }
}

// Find the symmetry operator and translation that moves
// the solution closest to the point and apply it.
// The point can be outside the canonical cell.
// The position along any polar axes is assumed to be already set to
// near the target point, so it isn't adjusted here.
void transform_segment_toward_point(rigid_body& seg, const unit_cell& cell,
                                    const std::array<double, 3>& point,
                                    std::array<double, 3>& fixed_origin) {
  // move the point into 0->1
  auto p = point;
  canonicalize(p);
  p = cell.orthogonalize(p);
  const auto& sg = cell.space_group();
  auto min_symop = size_t{0};
  auto min_distsq = numeric_limits<double>::max();
  auto min_origin = std::array<double, 3>{};
  auto norigins = fixed_origin[0] < -1 ? sg.origin_count() : 1;
  for (auto j = 0; j < norigins; ++j) {
    auto origin = fixed_origin[0] < -1 ? sg.get_origin(j) : fixed_origin;
    auto xyz = geom::sub(seg.position, origin);
    for (auto i = size_t{0}; i < sg.symmetry_operators().size(); ++i) {
      std::array<double, 3> sxyz =
          geom::transform(sg.symmetry_operators()[i], xyz);
      canonicalize(sxyz);
      auto distsq = geom::distsq(p, cell.orthogonalize(sxyz));
      if (distsq < min_distsq) {
        // save symm op + translation that gives closest distance
        min_symop = i;
        min_origin = origin;
        min_distsq = distsq;
      }
    }
  }
  // apply transformation
  // orthogonalize the symmetry operator: Ortho * Symm * OrthoInv
  const auto& ortho = cell.orthogonalization_matrix();
  const auto& frac = cell.fractionalization_matrix();
  const auto& s = sg.symmetry_operators()[min_symop];
  auto symm =
      std::array<std::array<double, 3>, 3>{{{{s[0][0], s[0][1], s[0][2]}},
                                            {{s[1][0], s[1][1], s[1][2]}},
                                            {{s[2][0], s[2][1], s[2][2]}}}};
  auto symm2 = std::array<std::array<double, 3>, 3>{};
  geom::mul(ortho, symm, symm2);
  geom::mul(symm2, frac, symm);
  auto symrot = geom::make_quat(symm);
  geom::normalize(symrot, symrot);
  geom::mulquat(symrot, seg.orientation, symrot);
  geom::normalize(symrot, seg.orientation);

  auto pos = geom::sub(seg.position, min_origin);
  pos = geom::transform(sg.symmetry_operators()[min_symop], pos);
  canonicalize(pos);
  // add any full unit cell translations necessary for original point
  pos[0] += floor(point[0]);
  pos[1] += floor(point[1]);
  pos[2] += floor(point[2]);
  seg.position = pos;
  fixed_origin = min_origin;
}

// Simple I to F conversion by the method of
// Sivia DS, & David WIF (1994) A Bayesian approach to extracting
// structure-factor
// amplitudes from powder diffraction data. Acta Crystallogr. A50, 703–714.
void int_to_sf_simple(struct reflection& ref) {
  auto iobs = ref.obs;
  auto sigiobs = ref.sigma;
  auto fobs = 0.5 * sqrt(2.0 * iobs +
                         sqrt(4.0 * iobs * iobs + 8.0 * sigiobs * sigiobs));
  ref.obs = static_cast<float>(fobs);
  ref.sigma = static_cast<float>(
      1.0 / sqrt((1.0 / (fobs * fobs)) +
                 (2.0 * (3.0 * fobs * fobs - iobs) / (sigiobs * sigiobs))));
}

void canonicalize(std::array<double, 3>& v) {
  geom::fmod(v, 1.0);
  if (v[0] < 0.0)
    v[0] += 1.0;
  if (v[1] < 0.0)
    v[1] += 1.0;
  if (v[2] < 0.0)
    v[2] += 1.0;
}

vector<reflection> load_reflections(const unit_cell& cell, double reslow,
                                    double reshigh, const string& path,
                                    string& mtz_column) {
  auto dataformat = string("phs");
  auto datatype = string("fobs");

  auto data = vector<reflection>{};
  auto r = reflection{};
  data.reserve(4000);
  if (path.rfind(".mtz") == path.size() - 4) {
    dataformat = "mtz";
  } else if (path.rfind(".sca") == path.size() - 4) {
    dataformat = "sca";
    datatype = "iobs";
  }
  if (dataformat == "mtz") {
    ifstream mtz(path, ios_base::in | ios_base::binary);
    if (mtz.fail()) {
      const auto errmsg = "unable to open the file ";
      throw ios_base::failure(errmsg + path);
    }
    mtz.exceptions(ios_base::badbit);
    if (!mtz_is_valid(mtz))
      throw runtime_error("the file is not a valid MTZ file");
    int columns[4];
    columns[0] = mtz_column_position(mtz, "H");
    columns[1] = mtz_column_position(mtz, "K");
    columns[2] = mtz_column_position(mtz, "L");
    auto label = mtz_column;
    // If there is a user-supplied label, use it.
    if (label != "") {
      auto pos = mtz_column_position(mtz, label);
      if (pos < 0) {
        throw runtime_error("column with label '" + label +
                            "' not found in MTZ file");
      }
    } else {
      // Otherwise, try the "standard" labels.
      const auto labels = {"FP", "F", "FNAT", "I", "IMEAN"};
      for (const auto l : labels) {
        auto pos = mtz_column_position(mtz, l);
        if (pos >= 0) {
          label = l;
          break;
        }
      }
    }
    if (label == "") {
      // Now assume the first structure factor column with a
      // common prefix will be the desired native data set.
      const auto labels = {"FP_", "F_", "I_"};
      for (const auto l : labels) {
        label = mtz_first_column_label_starting_with(mtz, l);
        if (label != "") {
          break;
        }
      }
    }
    if (label == "") {
      // Now just take the first one of the requisite type.
      const auto types = {'F', 'J', 'G', 'K'};
      for (const auto t : types) {
        label = mtz_first_column_of_type(mtz, t);
        if (label != "") {
          columns[3] = mtz_column_position(mtz, label);
          break;
        }
      }
    }
    if (label == "") {
      throw runtime_error("No suitable data column found in the MTZ file.\n"
                          "You can specify the column to be used by providing"
                          "the column label with the option '-Z'.");
    }
    columns[3] = mtz_column_position(mtz, label);
    switch (mtz_column_type(mtz, label)) {
      case 'F':
      case 'G':
        datatype = "fobs";
        break;
      case 'J':
      case 'K':
        datatype = "iobs";
        break;
      default:
        throw std::runtime_error("selected column in MTZ file (" + label +
                                 ") is not of a suitable type");
        break;
    }
    mtz_column = label;
    auto nref = mtz_nrefl(mtz);
    float fields[4];
    for (auto i = 0; i < nref; ++i) {
      mtz_data_columns(mtz, i, columns, fields, 4);
      // Missing reflections with values of NaN are passed over, but
      // isfinite() and isnan() are optimized away with options like gcc's
      // -ffast-math so we use range checks instead. Because these also
      // can fail for NaNs with certain compiler options, there is still an
      // isnan() check as well.
      auto fo = fields[3];
      if (fo >= 0.0f && fo < numeric_limits<float>::max() &&
          (!std::isnan(fo))) {
        r.hkl[0] = fields[0];
        r.hkl[1] = fields[1];
        r.hkl[2] = fields[2];
        r.obs = fo;
        auto res = cell.resolution(r.hkl[0], r.hkl[1], r.hkl[2]);
        if (res > reshigh && res < reslow) {
          data.push_back(r);
        }
      }
    }
  } else {
    ifstream datastream(path);
    if (datastream.fail()) {
      const auto errmsg = "unable to open the file ";
      throw ios_base::failure(errmsg + path);
    }
    datastream.exceptions(ios_base::badbit);
    // Scalepack file
    if (dataformat == "sca") {
      // skip header lines of scalepack file
      datastream.ignore(1000, '\n');
      datastream.ignore(1000, '\n');
      datastream.ignore(1000, '\n');
      auto sca_rec = string{};
      while (getline(datastream, sca_rec)) {
        if (sca_rec.length() < 28) {
          const auto errmsg = "invalid reflection data on this line: ";
          throw std::runtime_error(errmsg + sca_rec);
        }
        r.hkl[0] = stof(sca_rec.substr(0, 4));
        r.hkl[1] = stof(sca_rec.substr(4, 4));
        r.hkl[2] = stof(sca_rec.substr(8, 4));
        r.obs = stof(sca_rec.substr(12, 8));
        r.sigma = stof(sca_rec.substr(20, 8));
        if (sca_rec.length() >= 44) {
          auto ineg = stod(sca_rec.substr(28, 8));
          auto sigineg = stod(sca_rec.substr(36, 8));
          auto wineg = 1.0 / (sigineg * sigineg);
          auto wipos = 1.0 / (r.sigma * r.sigma);
          r.obs = float((r.obs * wipos + ineg * wineg) / (wipos + wineg));
          r.sigma = float(0.5 * sqrt(r.sigma * r.sigma + sigineg * sigineg));
        }
        auto res = cell.resolution(r.hkl[0], r.hkl[1], r.hkl[2]);
        if (res > reshigh && res < reslow) {
          data.push_back(r);
        }
      }
    }
    // generic, free-format h, k, l, fo and optional sigfo
    else {
      char hklrec[100];
      while (datastream.getline(hklrec, 100)) {
        auto nf = sscanf(hklrec, "%f %f %f %f %f", &r.hkl[0], &r.hkl[1],
                         &r.hkl[2], &r.obs, &r.sigma);
        if (nf < 4) {
          const auto errmsg = string{"invalid reflection data on this line: "};
          throw std::runtime_error(errmsg + hklrec);
        }
        if (nf < 5) {
          r.sigma = 1.0f;
        }
        auto res = cell.resolution(r.hkl[0], r.hkl[1], r.hkl[2]);
        if (res > reshigh && res < reslow) {
          data.push_back(r);
        }
      }
    }
  }
  if (datatype == "iobs") {
    for (auto& ref : data) {
      int_to_sf_simple(ref);
    }
  }
  return data;
}

static void create_alternate_cells(vector<unit_cell>& cells,
                                   vector<unsigned int>& permutations) {
  const auto& cell = cells[0];
  const auto& pointgroup = cell.space_group().point_group();
  const auto latticetype = cell.space_group().lattice_type();
  auto number = cell.space_group().number();
  for (auto i = 1; i <= spacegroup::NumberOfSpaceGroups; ++i) {
    auto& spgrp = spacegroup::from_number(i);
    if (spgrp.point_group() == pointgroup &&
        spgrp.lattice_type() == latticetype) {
      if (i != number) {
        cells.emplace_back(i, cells[0].a_axis(), cells[0].b_axis(),
                           cells[0].c_axis(), cells[0].alpha(), cells[0].beta(),
                           cells[0].gamma());
        permutations.push_back(0);
      }
      // do cell axis permutations when necessary
      // (including on originally designated space group)
      if (i == 17 || i == 18) {
        cells.emplace_back(i, cells[0].b_axis(), cells[0].c_axis(),
                           cells[0].a_axis(), cells[0].alpha(), cells[0].beta(),
                           cells[0].gamma());
        permutations.push_back(1);
        cells.emplace_back(i, cells[0].c_axis(), cells[0].a_axis(),
                           cells[0].b_axis(), cells[0].alpha(), cells[0].beta(),
                           cells[0].gamma());
        permutations.push_back(2);
      }
    }
  }
}

void report_cells(const vector<unit_cell>& cells, ostream& outs) {
  const auto Rad2Deg = 180.0 / geom::PI;
  for (const auto& cell : cells) {
    outs << "a=" << cell.a_axis() << " b=" << cell.b_axis()
         << " c=" << cell.c_axis() << " alpha=" << cell.alpha() * Rad2Deg
         << " beta=" << cell.beta() * Rad2Deg
         << " gamma=" << cell.gamma() * Rad2Deg << ", space group "
         << cell.space_group().number() << " ("
         << cell.space_group().hm_symbols()[0] << ") ";
    outs << '\n';
  }
  outs << endl;
}

void dsimplex_optimize(solution& sol,
                       scorer<pearson<float>, interpolation::Linear,
                              local_reflection_parallelism>& scorer,
                       double simplex_size, double simplex_tolerance,
                       bool rotate, bool translate, bool maximizing) {
  // move Ksol and Bsol into a compatible numeric range
  const auto bsol_divisor = 10000.0;
  const auto ksol_divisor = 10.0;
  auto dsimscore = [&](const vector<double>& solvec) {
    auto offset = 0;
    auto s = sol;
    for (auto& seg : s.segments) {
      if (rotate) {
        seg.orientation[0] = solvec[offset];
        seg.orientation[1] = solvec[offset + 1];
        seg.orientation[2] = solvec[offset + 2];
        seg.orientation[3] = solvec[offset + 3];
        geom::normalize(seg.orientation, seg.orientation);
        offset += 4;
      }
      // shift into the canonical cell for scoring to minimize the chance of
      // calculation errors
      if (translate) {
        seg.position[0] = fmod(solvec[offset], 1.0);
        seg.position[1] = fmod(solvec[offset + 1], 1.0);
        seg.position[2] = fmod(solvec[offset + 2], 1.0);
        offset += 3;
      }
    }
    s.ksol = static_cast<float>(ksol_divisor * solvec[offset++]);
    s.bsol = static_cast<float>(bsol_divisor * solvec[offset]);
    // dsimplex function only minimizes, so invert score if necessary
    return maximizing ? -scorer(s) : scorer(s);
  };
  auto paramvector = vector<double>{};
  for (const auto& seg : sol.segments) {
    if (rotate) {
      paramvector.push_back(seg.orientation[0]);
      paramvector.push_back(seg.orientation[1]);
      paramvector.push_back(seg.orientation[2]);
      paramvector.push_back(seg.orientation[3]);
    }
    if (translate) {
      paramvector.push_back(seg.position[0]);
      paramvector.push_back(seg.position[1]);
      paramvector.push_back(seg.position[2]);
    }
  }
  paramvector.push_back(sol.ksol / ksol_divisor);
  paramvector.push_back(sol.bsol / bsol_divisor);

  optimization::dsimplex(paramvector.begin(), paramvector.end(),
                         paramvector.begin(), dsimscore, simplex_size,
                         simplex_tolerance);
  auto offset = 0;
  for (auto& seg : sol.segments) {
    if (rotate) {
      seg.orientation[0] = paramvector[offset];
      seg.orientation[1] = paramvector[offset + 1];
      seg.orientation[2] = paramvector[offset + 2];
      seg.orientation[3] = paramvector[offset + 3];
      geom::normalize(seg.orientation, seg.orientation);
      offset += 4;
    }
    if (translate) {
      // don't do any position shifting here as it would mess up local searches
      // let the normalization routine handle it
      seg.position[0] = paramvector[offset];
      seg.position[1] = paramvector[offset + 1];
      seg.position[2] = paramvector[offset + 2];
      offset += 3;
    }
  }
  sol.ksol = static_cast<float>(ksol_divisor * paramvector[offset++]);
  sol.bsol = static_cast<float>(bsol_divisor * paramvector[offset]);
}
