#ifndef PDB_HPP
#define PDB_HPP

/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#include "atom.hpp"

#include <cctype>
#include <istream>
#include <string>

namespace pdb {

enum residue_class {
  Other = 0,
  StdAminoAcid = 1,
  NonStdAminoAcid = 2,
  AAMod = 4,
  AminoAcid = 7,
  StdNucleicAcid = 8,
  NonStdNucleicAcid = 16,
  NAMod = 32,
  NucleicAcid = 56,
  CommonCofactor = 64,
  Water = 128
};

// non-template functions are inline to allow header-only implementation

std::istream& operator>>(std::istream& s, atom& a);
int get_atomic_number(const char c1, const char c2, bool guess) noexcept;

// classify the residue the atom is in
// line can be string, char* or other indexable type
// does not do bounds checking
template <typename LineType>
residue_class get_residue_class(const LineType& line) noexcept {
  auto n = 17;
  while (line[n] == ' ')
    ++n;
  switch (line[n]) {
    case 'A':
      switch (line[n + 1]) {
        case 'B':
          if (line[n + 2] == 'U')
            return NonStdAminoAcid; /* gamma-Aminobutyric acid */
          break;
        case 'C':
          if (line[n + 2] == 'E')
            return AAMod;
          break;
        case 'L':
          if (line[n + 2] == 'A')
            return StdAminoAcid;
          break;
        case 'R':
          if (line[n + 2] == 'G')
            return StdAminoAcid;
          break;
        case 'S':
          if (line[n + 2] == 'N' || line[n + 2] == 'P' || line[n + 2] == 'X')
            return StdAminoAcid;
          break;
        case ' ':
          return StdNucleicAcid;
        default:
          break;
      };
      break;
    case 'C':
      if (line[n + 1] == 'Y' && line[n + 2] == 'S')
        return StdAminoAcid;
      else if (line[n + 1] == ' ')
        return StdNucleicAcid;
      break;
    case 'D':
      switch (line[n + 1]) {
        case 'A':
        case 'T':
        case 'G':
        case 'C':
        case 'U':
        case 'I':
          if (line[n + 2] == ' ')
            return StdNucleicAcid;
          break;
        case 'O':
          if (line[n + 2] == 'D')
            return Water;
          break;
        case '2':
          if (line[n + 2] == 'O')
            return Water;
          break;
        default:
          break;
      }
      break;
    case 'F':
      if (line[n + 1] == 'O' && line[n + 2] == 'R')
        return AAMod;
      break;
    case 'G':
      switch (line[n + 1]) {
        case 'L':
          if (line[n + 2] == 'N' || line[n + 2] == 'U' || line[n + 2] == 'Y' ||
              line[n + 2] == 'X')
            return StdAminoAcid;
          break;
        case ' ':
          return StdNucleicAcid;
        default:
          break;
      };
      break;
    case 'H':
      if ((line[n + 1] == 'I' && line[n + 2] == 'S') ||
          (line[n + 1] == 'Y' && line[n + 2] == 'L'))
        return StdAminoAcid;
      else if ((line[n + 1] == 'S' && line[n + 2] == 'E') ||
               (line[n + 1] == 'Y' && line[n + 2] == 'P'))
        return NonStdAminoAcid;
      else if ((line[n + 1] == 'O' && line[n + 2] == 'H') ||
               (line[n + 1] == '2' && line[n + 2] == 'O'))
        return Water;
      break;
    case 'I':
      if (line[n + 1] == 'L' && line[n + 2] == 'E')
        return StdAminoAcid;
      else if (line[n + 1] == ' ')
        return StdNucleicAcid;
      break;
    case 'L':
      if (line[n + 1] == 'E' && line[n + 2] == 'U')
        return StdAminoAcid;
      else if (line[n + 1] == 'Y' && line[n + 2] == 'S')
        return StdAminoAcid;
      break;
    case 'M':
      if (line[n + 1] == 'E' && line[n + 2] == 'T')
        return StdAminoAcid;
      if (line[n + 1] == 'S' && line[n + 2] == 'E')
        return NonStdAminoAcid;
      break;
    case 'N':
      if (line[n + 1] == 'H' && line[n + 2] == '2')
        return AAMod;
      break;
    case 'P':
      if (line[n + 1] == 'H' && line[n + 2] == 'E')
        return StdAminoAcid;
      else if (line[n + 1] == 'C' && line[n + 2] == 'A')
        return NonStdAminoAcid;
      else if (line[n + 1] == 'R' && line[n + 2] == 'O')
        return StdAminoAcid;
      break;
    case 'S':
      if (line[n + 1] == 'E' && line[n + 2] == 'R')
        return StdAminoAcid;
      break;
    case 'T':
      switch (line[n + 1]) {
        case 'H':
          if (line[n + 2] == 'R')
            return StdAminoAcid;
          break;
        case 'R':
          if (line[n + 2] == 'P')
            return StdAminoAcid;
          break;
        case 'Y':
          if (line[n + 2] == 'R')
            return StdAminoAcid;
        case ' ':
          return StdNucleicAcid;
        default:
          break;
      };
      break;
    case 'U':
      if (line[n + 1] == ' ')
        return StdNucleicAcid;
      break;
    case 'V':
      if (line[n + 1] == 'A' && line[n + 2] == 'L')
        return StdAminoAcid;
      break;
    case 'W':
      if (line[n + 1] == 'A' && line[n + 2] == 'T')
        return Water;
      break;
    case '+':
      switch (line[n + 1]) {
        case 'A':
        case 'T':
        case 'G':
        case 'C':
        case 'U':
        case 'I':
          if (line[n + 2] == ' ')
            return NonStdNucleicAcid;
          break;
        case 'D':
          switch (line[n + 2]) {
            case 'A':
            case 'T':
            case 'G':
            case 'C':
            case 'U':
            case 'I':
              if (line[n + 3] == ' ')
                return NonStdNucleicAcid;
              break;
            default:
              break;
          }
          break;
        default:
          break;
      }
      break;
    default:
      break;
  }
  return Other;
}

inline int get_atomic_number(const char c1, const char c2,
                             bool guess) noexcept {
  if (c1 == ' ' || (guess && (!isalpha(c1)))) {
    switch (c2) {
      case 'C':
        return 6;
      case 'N':
        return 7;
      case 'O':
        return 8;
      case 'S':
        return 16;
      case 'P':
        return 15;
      case 'H':
        return 1;
      case 'I':
        return 53;
      case 'F':
        return 9;
      case 'U':
        return 92;
      case 'K':
        return 19;
      case 'W':
        return 74;
      case 'V':
        return 23;
      case 'B':
        return 5;
      case 'Y':
        return 39;
      case 'D':
        return 1;
      /* PDB files use X to indicate an unknown atom */
      case 'X':
        return 0;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'A') {
    switch (c2) {
      case 'L':
        return 13;
      case 'R':
        return 18;
      case 'S':
        return 33;
      case 'G':
        return 47;
      case 'U':
        return 79;
      case 'T':
        return 85;
      case 'C':
        return 89;
      case 'M':
        return 95;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'B') {
    switch (c2) {
      case 'A':
        return 56;
      case 'H':
        return 107;
      case 'I':
        return 83;
      case 'R':
        return 35;
      case 'E':
        return 4;
      case 'K':
        return 97;
      default:
        return guess ? 5 : -1;
    }
  } else if (c1 == 'C') {
    switch (c2) {
      case 'L':
        return 17;
      case 'A':
        return 20;
      case 'R':
        return 24;
      case 'O':
        return 27;
      case 'U':
        return 29;
      case 'D':
        return 48;
      case 'S':
        return 55;
      case 'E':
        return 58;
      case 'M':
        return 96;
      case 'F':
        return 98;
      case 'N':
        return 112;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'D') {
    switch (c2) {
      case 'B':
        return 105;
      case 'S':
        return 110;
      case 'Y':
        return 66;
      default:
        return guess ? 1 : -1;
    }
  } else if (c1 == 'E') {
    switch (c2) {
      case 'U':
        return 63;
      case 'R':
        return 68;
      case 'S':
        return 99;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'F') {
    switch (c2) {
      case 'E':
        return 26;
      case 'M':
        return 100;
      case 'R':
        return 87;
      default:
        return guess ? 9 : -1;
    }
  } else if (c1 == 'G') {
    switch (c2) {
      case 'A':
        return 31;
      case 'E':
        return 32;
      case 'D':
        return 64;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'H') {
    switch (c2) {
      case 'E':
        return 2;
      case 'F':
        return 72;
      case 'G':
        return 80;
      case 'S':
        return 108;
      case 'O':
        return 67;
      default:
        return guess ? 1 : -1;
    }
  } else if (c1 == 'I') {
    switch (c2) {
      case 'N':
        return 49;
      case 'R':
        return 77;
      default:
        return guess ? 53 : -1;
    }
  } else if (c1 == 'K') {
    switch (c2) {
      case 'R':
        return 36;
      default:
        return guess ? 19 : -1;
    }
  } else if (c1 == 'L') {
    switch (c2) {
      case 'I':
        return 3;
      case 'A':
        return 57;
      case 'R':
        return 103;
      case 'U':
        return 71;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'M') {
    switch (c2) {
      case 'D':
        return 101;
      case 'G':
        return 12;
      case 'N':
        return 25;
      case 'O':
        return 42;
      case 'T':
        return 109;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'N') {
    switch (c2) {
      case 'A':
        return 11;
      case 'B':
        return 41;
      case 'D':
        return 60;
      case 'E':
        return 10;
      case 'I':
        return 28;
      case 'O':
        return 102;
      case 'P':
        return 93;
      default:
        return guess ? 7 : -1;
    }
  } else if (c1 == 'O') {
    switch (c2) {
      case 'S':
        return 76;
      default:
        return guess ? 8 : -1;
    }
  } else if (c1 == 'P') {
    switch (c2) {
      case 'A':
        return 91;
      case 'B':
        return 82;
      case 'D':
        return 46;
      case 'M':
        return 61;
      case 'O':
        return 84;
      case 'R':
        return 59;
      case 'T':
        return 78;
      case 'U':
        return 94;
      default:
        return guess ? 15 : -1;
    }
  } else if (c1 == 'R') {
    switch (c2) {
      case 'A':
        return 88;
      case 'B':
        return 37;
      case 'E':
        return 75;
      case 'F':
        return 104;
      case 'G':
        return 111;
      case 'H':
        return 45;
      case 'N':
        return 86;
      case 'U':
        return 44;
      default:
        return guess ? 6 : -1;
    }
  } else if (c1 == 'S') {
    switch (c2) {
      case 'E':
        return 34;
      case 'I':
        return 14;
      case 'C':
        return 21;
      case 'R':
        return 38;
      case 'N':
        return 50;
      case 'B':
        return 51;
      case 'M':
        return 62;
      case 'G':
        return 106;
      default:
        return guess ? 16 : -1;
    }
  } else if (c1 == 'T') {
    switch (c2) {
      case 'A':
        return 73;
      case 'B':
        return 65;
      case 'C':
        return 43;
      case 'E':
        return 52;
      case 'H':
        return 90;
      case 'I':
        return 22;
      case 'L':
        return 81;
      case 'M':
        return 69;
      default:
        return guess ? 6 : -1;
    }
  } else if (guess && c1 == 'U') {
    /* Since we're guessing, UN is more likely to be the start of UNK */
    return c2 == 'N' ? 0 : 92;
  } else if (c1 == 'V') {
    return guess ? 23 : -1;
  } else if (c1 == 'W') {
    return guess ? 74 : -1;
  } else if (c1 == 'X') {
    if (c2 == 'E')
      return 54;
    else
      return guess ? 0 : -1;
  } else if (c1 == 'Y') {
    if (c2 == 'B')
      return 70;
    else
      return guess ? 39 : -1;
  } else if (c1 == 'Z') {
    switch (c2) {
      case 'N':
        return 30;
      case 'R':
        return 40;
      default:
        return guess ? 6 : -1;
    }
  }
  /* default is carbon if guessing */
  if (guess) {
    switch (c2) {
      case 'C':
        return 6;
      case 'N':
        return 7;
      case 'O':
        return 8;
      case 'P':
        return 15;
      case 'S':
        return 16;
      default:
        return 6;
    }
  } else
    return -1;
}

template <typename StringType>
int get_atomic_number(const StringType& line, size_t len) noexcept {
  auto element = len < 78 ? -1 : get_atomic_number(line[76], line[77], false);
  if (element < 0) {
    element = get_atomic_number(line[12], line[13], true);
    // extra check for possible mis-read hydrogens
    // (if left-shifted name and no element code)
    if (line[12] == 'H') {
      auto c = get_residue_class(line);
      if ((c & AminoAcid) || (c & NucleicAcid))
        element = 1;
    }
  }
  if (element < 1)
    element = 6; // set to carbon as a last resort
  return element;
}

inline std::istream& operator>>(std::istream& s, atom& a) {
  // skip lines until an atom record is found or eof is reached
  auto len = size_t{0};
  do {
    if (!getline(s, a.line))
      return s;
  } while ((a.line.compare(0, 6, "ATOM  ") != 0 &&
            a.line.compare(0, 6, "HETATM") != 0) ||
           (len = a.line.length()) < 53);
  a.position = {{std::stod(a.line.substr(30, 8)),
                 std::stod(a.line.substr(38, 8)),
                 std::stod(a.line.substr(46, 8))}};
  if (len >= 59) {
    a.occ = std::stod(a.line.substr(54, 6));
  } else
    a.occ = 1.0;
  if (len >= 65) {
    a.b = std::stod(a.line.substr(60, 6));
  } else
    a.b = 0.0;
  a.Z = static_cast<short>(get_atomic_number(a.line, len));
  return s;
}
}
#endif
