/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef CALCFC_HPP
#define CALCFC_HPP

#include "moleculartransform.hpp"
#include "platform.hpp"
#include "reflection.hpp"
#include "rfactor.hpp"
#include "solution.hpp"
#include "statistics.hpp"

#include <algorithm>
#include <complex>
#include <future>
#include <memory>
#include <vector>

using std::complex;
using std::vector;
using crystallography::unit_cell;
using crystallography::molecular_transform;

enum class interpolation { None, Linear };

template <interpolation Interp>
inline complex<float> get_grid_value(const grid3<complex<float>>& grid, float h,
                                     float k, float l) {
  return grid.nearest_value<float>(h, k, l);
}

template <>
inline complex<float>
get_grid_value<interpolation::Linear>(const grid3<complex<float>>& grid,
                                      float h, float k, float l) {
  return grid.linear_interpolated_value(h, k, l);
}

class fc_calc_base {
public:
  fc_calc_base(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : transf(transforms), cells(unitcells),
        hkl_permutations(index_permutations) {
    for (const auto& r : reflections) {
      stolsq.emplace_back(r.stolsq);
      hkl.emplace_back(r.hkl);
      fpart.emplace_back(r.fpartial);
      fobs.emplace_back(r.obs);
      weights.emplace_back(
          r.sigma); // r.sigma holds a weight rather than sigma at this point
    };
  };

protected:
  vector<float> stolsq;
  vector<std::array<float, 3>> hkl;
  vector<complex<float>> fpart;
  vector<float> fobs;
  vector<float> weights;
  const vector<std::unique_ptr<molecular_transform<float>>>& transf;
  const vector<unit_cell>& cells;
  const vector<unsigned int>& hkl_permutations;
};

template <interpolation Interp, parallelism ParallelType>
class fc_calculator;

template <interpolation Interp>
class fc_calculator<Interp, parallelism::None> : protected fc_calc_base {
public:
  fc_calculator(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : fc_calc_base(reflections, transforms, unitcells, index_permutations) {}
  void
  calcfc(vector<complex<float>>& fc, const grid3<complex<float>>& grid,
         const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
         const std::array<double, 3>& trans,
         const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
         const unsigned int h_ind, const unsigned int k_ind,
         const unsigned int l_ind) const;
};

template <interpolation Interp>
void fc_calculator<Interp, parallelism::None>::calcfc(
    vector<complex<float>>& fc, const grid3<complex<float>>& grid,
    const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
    const std::array<double, 3>& trans,
    const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
    const unsigned int h_ind, const unsigned int k_ind,
    const unsigned int l_ind) const {
  const auto nref = fc.size();
  const auto nsymm = symmops.size();
  for (auto i = size_t{0}; i < nref; ++i) {
    const auto h = this->hkl[i][h_ind];
    const auto k = this->hkl[i][k_ind];
    const auto l = this->hkl[i][l_ind];
    auto fcalc = complex<float>{};
    for (auto is = size_t{0}; is < nsymm; ++is) {
      const auto& c2g = cell2grid[is];
      const auto gridh = float(h * c2g[0][0] + k * c2g[0][1] + l * c2g[0][2]);
      const auto gridk = float(h * c2g[1][0] + k * c2g[1][1] + l * c2g[1][2]);
      const auto gridl = float(h * c2g[2][0] + k * c2g[2][1] + l * c2g[2][2]);
      const auto sf =
          gridl < 0 ? conj(get_grid_value<Interp>(grid, -gridh, -gridk, -gridl))
                    : get_grid_value<Interp>(grid, gridh, gridk, gridl);
      const auto& symm = symmops[is];
      // multiply by transpose of symm op
      auto dph = trans[0] * (h * symm[0][0] + k * symm[1][0] + l * symm[2][0]);
      dph += trans[1] * (h * symm[0][1] + k * symm[1][1] + l * symm[2][1]);
      dph += trans[2] * (h * symm[0][2] + k * symm[1][2] + l * symm[2][2]);
      dph += h * symm[0][3] + k * symm[1][3] + l * symm[2][3];
      dph *= 2.0 * geom::PI;
      const auto fph = float(dph);
      const auto cosfc = std::cos(fph);
      const auto sinfc = std::sin(fph);
      fcalc += complex<float>{real(sf) * cosfc - imag(sf) * sinfc,
                              real(sf) * sinfc + imag(sf) * cosfc};
    }
    fc[i] += fcalc;
  }
}

template <interpolation Interp>
class fc_calculator<Interp, parallelism::OpenMP> : protected fc_calc_base {
public:
  fc_calculator(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : fc_calc_base(reflections, transforms, unitcells, index_permutations) {}
  void
  calcfc(vector<complex<float>>& fc, const grid3<complex<float>>& grid,
         const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
         const std::array<double, 3>& trans,
         const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
         const unsigned int h_ind, const unsigned int k_ind,
         const unsigned int l_ind) const;
};

template <interpolation Interp>
void fc_calculator<Interp, parallelism::OpenMP>::calcfc(
    vector<complex<float>>& fc, const grid3<complex<float>>& grid,
    const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
    const std::array<double, 3>& trans,
    const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
    const unsigned int h_ind, const unsigned int k_ind,
    const unsigned int l_ind) const {
  const auto nref = fc.size();
  const auto nsymm = symmops.size();
#pragma omp parallel for
  for (auto i = size_t{0}; i < nref; ++i) {
    const auto h = this->hkl[i][h_ind];
    const auto k = this->hkl[i][k_ind];
    const auto l = this->hkl[i][l_ind];
    auto fcalc = complex<float>{};
    for (auto is = size_t{0}; is < nsymm; ++is) {
      const auto& c2g = cell2grid[is];
      const auto gridh = float(h * c2g[0][0] + k * c2g[0][1] + l * c2g[0][2]);
      const auto gridk = float(h * c2g[1][0] + k * c2g[1][1] + l * c2g[1][2]);
      const auto gridl = float(h * c2g[2][0] + k * c2g[2][1] + l * c2g[2][2]);
      const auto sf =
          gridl < 0 ? conj(get_grid_value<Interp>(grid, -gridh, -gridk, -gridl))
                    : get_grid_value<Interp>(grid, gridh, gridk, gridl);
      const auto& symm = symmops[is];
      // multiply by transpose of symm op
      auto dph = trans[0] * (h * symm[0][0] + k * symm[1][0] + l * symm[2][0]);
      dph += trans[1] * (h * symm[0][1] + k * symm[1][1] + l * symm[2][1]);
      dph += trans[2] * (h * symm[0][2] + k * symm[1][2] + l * symm[2][2]);
      dph += h * symm[0][3] + k * symm[1][3] + l * symm[2][3];
      dph *= 2.0 * geom::PI;
      const auto fph = float(dph);
      const auto cosfc = std::cos(fph);
      const auto sinfc = std::sin(fph);
      fcalc += complex<float>{real(sf) * cosfc - imag(sf) * sinfc,
                              real(sf) * sinfc + imag(sf) * cosfc};
    }
    fc[i] += fcalc;
  }
}

#if defined(__APPLE__) && defined(__BLOCKS__)
#include <dispatch/dispatch.h>

template <interpolation Interp>
class fc_calculator<Interp, parallelism::GCD> : protected fc_calc_base {
public:
  fc_calculator(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : fc_calc_base(reflections, transforms, unitcells, index_permutations) {
    queue = dispatch_get_global_queue(0, 0);
  }

  void
  calcfc(vector<complex<float>>& fc, const grid3<complex<float>>& grid,
         const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
         const std::array<double, 3>& trans,
         const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
         const unsigned int h_ind, const unsigned int k_ind,
         const unsigned int l_ind) const;

protected:
  dispatch_queue_t queue;
};

template <interpolation Interp>
void fc_calculator<Interp, parallelism::GCD>::calcfc(
    vector<complex<float>>& fc, const grid3<complex<float>>& grid,
    const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
    const std::array<double, 3>& trans,
    const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
    const unsigned int h_ind, const unsigned int k_ind,
    const unsigned int l_ind) const {
  const auto nref = fc.size();
  const auto nsymm = symmops.size();
  dispatch_apply(nref, queue, ^(size_t i) {
    const auto h = this->hkl[i][h_ind];
    const auto k = this->hkl[i][k_ind];
    const auto l = this->hkl[i][l_ind];
    auto fcalc = complex<float>{};
    for (auto is = size_t{0}; is < nsymm; ++is) {
      const auto& c2g = cell2grid[is];
      const auto gridh = float(h * c2g[0][0] + k * c2g[0][1] + l * c2g[0][2]);
      const auto gridk = float(h * c2g[1][0] + k * c2g[1][1] + l * c2g[1][2]);
      const auto gridl = float(h * c2g[2][0] + k * c2g[2][1] + l * c2g[2][2]);
      const auto sf =
          gridl < 0 ? conj(get_grid_value<Interp>(grid, -gridh, -gridk, -gridl))
                    : get_grid_value<Interp>(grid, gridh, gridk, gridl);
      const auto& symm = symmops[is];
      // multiply by transpose of symm op
      auto dph = trans[0] * (h * symm[0][0] + k * symm[1][0] + l * symm[2][0]);
      dph += trans[1] * (h * symm[0][1] + k * symm[1][1] + l * symm[2][1]);
      dph += trans[2] * (h * symm[0][2] + k * symm[1][2] + l * symm[2][2]);
      dph += h * symm[0][3] + k * symm[1][3] + l * symm[2][3];
      dph *= 2.0 * geom::PI;
      const auto fph = float(dph);
      const auto cosfc = std::cos(fph);
      const auto sinfc = std::sin(fph);
      fcalc += complex<float>{real(sf) * cosfc - imag(sf) * sinfc,
                              real(sf) * sinfc + imag(sf) * cosfc};
    }
    fc[i] += fcalc;
  });
}

#endif

template <interpolation Interp>
class fc_calculator<Interp, parallelism::PPL> : protected fc_calc_base {
public:
  fc_calculator(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : fc_calc_base(reflections, transforms, unitcells, index_permutations) {}
  void
  calcfc(vector<complex<float>>& fc, const grid3<complex<float>>& grid,
         const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
         const std::array<double, 3>& trans,
         const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
         const unsigned int h_ind, const unsigned int k_ind,
         const unsigned int l_ind) const;
};

#ifdef _MSC_VER
#include <ppl.h>

template <interpolation Interp>
void fc_calculator<Interp, parallelism::PPL>::calcfc(
    vector<complex<float>>& fc, const grid3<complex<float>>& grid,
    const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
    const std::array<double, 3>& trans,
    const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
    const unsigned int h_ind, const unsigned int k_ind,
    const unsigned int l_ind) const {
  const auto nref = fc.size();
  const auto nsymm = symmops.size();
  Concurrency::parallel_for(size_t{0}, nref, [&](size_t i) {
    const auto h = this->hkl[i][h_ind];
    const auto k = this->hkl[i][k_ind];
    const auto l = this->hkl[i][l_ind];
    auto fcalc = complex<float>{};
    for (auto is = size_t{0}; is < nsymm; ++is) {
      const auto& c2g = cell2grid[is];
      const auto gridh = float(h * c2g[0][0] + k * c2g[0][1] + l * c2g[0][2]);
      const auto gridk = float(h * c2g[1][0] + k * c2g[1][1] + l * c2g[1][2]);
      const auto gridl = float(h * c2g[2][0] + k * c2g[2][1] + l * c2g[2][2]);
      const auto sf =
          gridl < 0 ? conj(get_grid_value<Interp>(grid, -gridh, -gridk, -gridl))
                    : get_grid_value<Interp>(grid, gridh, gridk, gridl);
      const auto& symm = symmops[is];
      // multiply by transpose of symm op
      auto dph = trans[0] * (h * symm[0][0] + k * symm[1][0] + l * symm[2][0]);
      dph += trans[1] * (h * symm[0][1] + k * symm[1][1] + l * symm[2][1]);
      dph += trans[2] * (h * symm[0][2] + k * symm[1][2] + l * symm[2][2]);
      dph += h * symm[0][3] + k * symm[1][3] + l * symm[2][3];
      dph *= 2.0 * geom::PI;
      const auto fph = float(dph);
      const auto cosfc = std::cos(fph);
      const auto sinfc = std::sin(fph);
      fcalc += complex<float>{real(sf) * cosfc - imag(sf) * sinfc,
                              real(sf) * sinfc + imag(sf) * cosfc};
    }
    fc[i] += fcalc;
  });
}

#endif

template <interpolation Interp>
class fc_calculator<Interp, parallelism::Async> : protected fc_calc_base {
public:
  fc_calculator(
      const vector<reflection>& reflections,
      const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
      const vector<unit_cell>& unitcells,
      const vector<unsigned int>& index_permutations)
      : fc_calc_base(reflections, transforms, unitcells, index_permutations) {
    nthr = std::max(1u, std::thread::hardware_concurrency());
  }
  void
  calcfc(vector<complex<float>>& fc, const grid3<complex<float>>& grid,
         const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
         const std::array<double, 3>& trans,
         const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
         const unsigned int h_ind, const unsigned int k_ind,
         const unsigned int l_ind) const;

protected:
  long nthr;
};

template <interpolation Interp>
void fc_calculator<Interp, parallelism::Async>::calcfc(
    vector<complex<float>>& fc, const grid3<complex<float>>& grid,
    const std::array<std::array<std::array<double, 3>, 3>, 192>& cell2grid,
    const std::array<double, 3>& trans,
    const std::vector<std::array<std::array<double, 4>, 4>>& symmops,
    const unsigned int h_ind, const unsigned int k_ind,
    const unsigned int l_ind) const {
  const auto nref = fc.size();
  auto nsymm = symmops.size();
  auto stride = std::max(size_t{1}, size_t{nref / nthr});
  auto first = size_t{0};
  std::vector<std::future<void>> futures;
  while (first < nref) {
    const auto limit = std::min(nref - first, stride) + first;
    futures.emplace_back(
        std::async(std::launch::async, [&, first, limit, this]() {
          for (auto i = first; i < limit; ++i) {
            const auto h = this->hkl[i][h_ind];
            const auto k = this->hkl[i][k_ind];
            const auto l = this->hkl[i][l_ind];
            auto fcalc = complex<float>{};
            for (auto is = size_t{0}; is < nsymm; ++is) {
              const auto& c2g = cell2grid[is];
              const auto gridh =
                  float(h * c2g[0][0] + k * c2g[0][1] + l * c2g[0][2]);
              const auto gridk =
                  float(h * c2g[1][0] + k * c2g[1][1] + l * c2g[1][2]);
              const auto gridl =
                  float(h * c2g[2][0] + k * c2g[2][1] + l * c2g[2][2]);
              const auto sf =
                  gridl < 0 ? conj(get_grid_value<Interp>(grid, -gridh, -gridk,
                                                          -gridl))
                            : get_grid_value<Interp>(grid, gridh, gridk, gridl);
              const auto& symm = symmops[is];
              // multiply by transpose of symm op
              auto dph =
                  trans[0] * (h * symm[0][0] + k * symm[1][0] + l * symm[2][0]);
              dph +=
                  trans[1] * (h * symm[0][1] + k * symm[1][1] + l * symm[2][1]);
              dph +=
                  trans[2] * (h * symm[0][2] + k * symm[1][2] + l * symm[2][2]);
              dph += h * symm[0][3] + k * symm[1][3] + l * symm[2][3];
              dph *= 2.0 * geom::PI;
              const auto fph = float(dph);
              const auto cosfc = std::cos(fph);
              const auto sinfc = std::sin(fph);
              fcalc += complex<float>{real(sf) * cosfc - imag(sf) * sinfc,
                                      real(sf) * sinfc + imag(sf) * cosfc};
            }
            fc[i] += fcalc;
          }
        }));
    first = limit;
  }
  for (auto& fut : futures) {
    fut.wait();
  }
}

template <class Stat, interpolation Interp, parallelism ParallelType>
class scorer : fc_calculator<Interp, ParallelType> {
public:
  scorer(const vector<reflection>& reflections,
         const vector<std::unique_ptr<molecular_transform<float>>>& transforms,
         const vector<unit_cell>& unitcells,
         const vector<unsigned int>& index_permutations)
      : fc_calculator<Interp, ParallelType>(reflections, transforms, unitcells,
                                            index_permutations) {}

  double operator()(const solution& s) const {
    auto stat = Stat();
    auto fc = this->calc_reflections(s);
    for (auto i = size_t{0}; i < fc.size(); ++i) {
      // sqrt(a^2 + b^2) is used rather than abs(std::complex) as it is
      // safe in this case and gives a significant speed up on VC++
      // stat.add(this->fobs[i], std::abs(fc[i]), this->weights[i]);
      stat.add(this->fobs[i], std::sqrt(fc[i].real() * fc[i].real() +
                                        fc[i].imag() * fc[i].imag()),
               this->weights[i]);
    }
    return stat();
  }

private:
  vector<complex<float>> calc_reflections(const solution& s) const;
};

template <class Stat, interpolation Interp, parallelism ParallelType>
vector<complex<float>>
scorer<Stat, Interp, ParallelType>::calc_reflections(const solution& s) const {
  const auto nref = this->fpart.size();
  const auto& symmops =
      this->cells[s.target_id].space_group().symmetry_operators();
  vector<complex<float>> fc(this->fpart);
  auto cell2grid = std::array<std::array<std::array<double, 3>, 3>, 192>();
  const auto h_ind = this->hkl_permutations[s.target_id];
  const auto k_ind = (this->hkl_permutations[s.target_id] + 1) % 3;
  const auto l_ind = (this->hkl_permutations[s.target_id] + 2) % 3;
  for (const auto& seg : s.segments) {
    const auto& transform = *this->transf[seg.model_id];
    const auto& grid = transform.reflection_grid();
    const auto trans = seg.position;
    auto mat = geom::make_matrix33(seg.orientation);
    auto mat2 = std::array<std::array<double, 3>, 3>();
    geom::mul(mat, transform.unitcell().orthogonalization_matrix(), mat2);
    geom::mul(this->cells[s.target_id].fractionalization_matrix(), mat2, mat);
    geom::transpose(mat);
    const auto nsymm = symmops.size();
    for (auto i = size_t{0}; i < nsymm; ++i) {
      auto s = geom::make_matrix33(symmops[i]);
      geom::mul_transpose_right(mat, s, cell2grid[i]);
    }
    this->calcfc(fc, grid, cell2grid, trans, symmops, h_ind, k_ind, l_ind);
  }
  for (auto i = size_t{0}; i < nref; ++i) {
    fc[i] *= 1.0f - s.ksol * std::exp(-s.bsol * this->stolsq[i]);
  }
  return fc;
}

#endif
