/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef STATISTICS_HPP
#define STATISTICS_HPP

// Statistics.hpp
// contains template functors and function calls for sum, mean, sample variance,
// population variance, linear correlation coefficient
// To get standard deviation: sqrt(variance())

#include <cmath>
#include <algorithm>
#include <limits>

// calculates weighted or unweighted Pearson correlation coefficient
template <class T, class ResultT = double>
class pearson {
public:
  pearson(){};
  void add(T a, T b, T weight = 1);
  ResultT operator()() const;
  void reset();

private:
  size_t n = 0;
  ResultT meana = 0;
  ResultT meanb = 0;
  ResultT vara = 0;
  ResultT varb = 0;
  ResultT weight_sum = 0;
  ResultT r = 0;
};

template <class T, class ResultT>
void pearson<T, ResultT>::add(T a, T b, T weight) {
  auto ai = static_cast<ResultT>(a);
  auto bi = static_cast<ResultT>(b);
  auto wi = static_cast<ResultT>(weight);
  ++n;
  auto temp = weight_sum + wi;
  auto delta_a = ai - meana;
  auto scale_delta_a = delta_a * wi / temp;
  vara += scale_delta_a * delta_a * weight_sum;
  meana += scale_delta_a;
  auto delta_b = bi - meanb;
  auto scale_delta_b = delta_b * wi / temp;
  meanb += scale_delta_b;
  varb += scale_delta_b * delta_b * weight_sum;
  weight_sum = temp;
  r += delta_a * delta_b * (n - 1) * wi / n;
}

template <class T, class ResultT>
inline ResultT pearson<T, ResultT>::operator()() const {
  return r / std::sqrt(vara * varb);
}

template <class T, class ResultT>
void pearson<T, ResultT>::reset() {
  n = 0;
  meana = 0;
  meanb = 0;
  vara = 0;
  varb = 0;
  weight_sum = 0;
  r = 0;
}

template <class In1, class In2, class In3, class ResultT = double>
ResultT calc_pearson(In1 aFirst, In1 aLast, In2 bFirst, In3 weightFirst) {
  pearson<typename std::iterator_traits<In1>::value_type, ResultT> cc;
  while (aFirst != aLast)
    cc.add(*aFirst++, *bFirst++, *weightFirst++);
  return cc();
}

template <class In1, class In2, class In3, class ResultT = double>
ResultT calc_pearson(In1 aFirst, In1 aLast, In2 bFirst) {
  pearson<typename std::iterator_traits<In1>::value_type, ResultT> cc;
  while (aFirst != aLast)
    cc.add(*aFirst++, *bFirst++);
  return cc();
}

template <class T>
class sum {
public:
  sum() {}
  void add(T x) { s += x; }
  T operator()() const { return s; }
  void reset() { s = 0.0; }

private:
  T s = 0;
};

template <class In>
typename std::iterator_traits<In>::value_type calc_sum(In first, In last) {
  sum<typename std::iterator_traits<In>::value_type> s;
  while (first != last)
    s(*first++);
  return s();
}

template <class T>
class maximum {
public:
  maximum() : maxim(std::numeric_limits<T>::lowest()) {}
  void add(T x) {
    if (x > maxim)
      maxim = x;
  }
  T operator()() const { return maxim; }
  void reset() { maxim = std::numeric_limits<T>::lowest(); }

private:
  T maxim;
};

template <class T>
class minimum {
public:
  minimum() : minim(std::numeric_limits<T>::max()) {}
  void add(T x) {
    if (x < minim)
      minim = x;
  }
  T operator()() const { return minim; }
  void reset() { minim = std::numeric_limits<T>::max(); }

private:
  T minim;
};

template <class T, class ResultT = double>
class mean {
public:
  mean(){};
  void add(T x);
  ResultT operator()() const;
  void reset();

private:
  ResultT m = 0;
  size_t n = 0;
};

template <class T, class ResultT>
inline void mean<T, ResultT>::reset() {
  m = 0.0;
  n = 0;
}

template <class T, class ResultT>
inline void mean<T, ResultT>::add(T x) {
  m += (static_cast<ResultT>(x) - m) / ++n;
}

template <class T, class ResultT>
inline ResultT mean<T, ResultT>::operator()() const {
  return m;
}

template <class In, class ResultT = double>
ResultT calc_mean(In first, In last) {
  mean<typename std::iterator_traits<In>::value_type> m;
  while (first != last)
    m.add(*first++);
  return m();
}

template <class T, class ResultT = double>
class population_variance {
public:
  population_variance(){};
  void add(T x);
  virtual ResultT operator()() const;
  void reset();

protected:
  ResultT mean = 0;
  ResultT m2 = 0;
  size_t n = 0;
};

template <class T, class ResultT>
inline void population_variance<T, ResultT>::add(T x) {
  ++n;
  ResultT delta = x - mean;
  mean += delta / n;
  m2 += delta * (x - mean);
}

template <class T, class ResultT>
ResultT population_variance<T, ResultT>::operator()() const {
  return m2 / n;
}

template <class T, class ResultT>
inline void population_variance<T, ResultT>::reset() {
  mean = 0;
  m2 = 0;
  n = 0;
}

template <class In, class ResultT = double>
ResultT calc_population_variance(In first, In last) {
  population_variance<typename std::iterator_traits<In>::value_type, ResultT>
      var;
  while (first != last)
    var.add(*first++);
  return var();
}

template <class T, class ResultT = double>
class sample_variance : public population_variance<T, ResultT> {
public:
  ResultT operator()() const override { return this->m2 / (this->n - 1); }
};

template <class In, class ResultT = double>
ResultT calc_sample_variance(In first, In last) {
  sample_variance<typename std::iterator_traits<In>::value_type, ResultT> var;
  while (first != last)
    var.add(*first++);
  return var();
}

#endif
