/*
 Copyright (c) 2015 by Charles R. Kissinger */

#ifndef UNITCELL_HPP
#define UNITCELL_HPP

#include "spacegroup.hpp"
#include "geometry.hpp"
#include <limits>
#include <array>
#include <cmath>

namespace crystallography {

// Encapsulates cell constant and symmetry information for a crystal
class unit_cell {

public:
  // Creates a unit_cell from the cell constants and space group number.
  // The angles are assumed to be specified in radians if their values
  // are less than pi, and in degrees otherwise.
  unit_cell(int space_group_number, double a, double b, double c, double alpha,
            double beta, double gamma);

  // Returns the a axis length in Angstroms
  double a_axis() const { return a; }
  // Returns the b axis length in Angstroms
  double b_axis() const { return b; }
  // Returns the c axis length in Angstroms
  double c_axis() const { return c; }
  // Returns the alpha angle in radians
  double alpha() const { return alph; }
  // Returns the beta angle in radians
  double beta() const { return bet; }
  // Returns the gamma angle in radians
  double gamma() const { return gam; }
  // Returns the unit cell volume in cubic Angstroms
  double volume() const { return vol; }
  // Returns the a* axis length in reciprocal Angstroms
  double recip_a() const { return ra; }
  // Returns the b* axis length in reciprocal Angstroms
  double recip_b() const { return rb; }
  // Returns the c* axis length in reciprocal Angstroms
  double recip_c() const { return rc; }
  // Returns the alpha* angle in radians
  double recip_alpha() const { return ralpha; };
  // Returns the beta* angle in radians
  double recip_beta() const { return rbeta; };
  // Returns the beta* angle in radians
  double recip_gamma() const { return rgamma; };

  // Returns a reference to a PDB-standard upper-triangular orthogonalization
  // matrix for the cell
  const std::array<std::array<double, 3>, 3>& orthogonalization_matrix() const {
    return frac_to_cart;
  };

  // Returns a reference to an upper-triangular fractionalization matrix for the
  // cell
  const std::array<std::array<double, 3>, 3>& fractionalization_matrix() const {
    return cart_to_frac;
  };

  // Returns the fractional coordinates corresponding to the supplied orthogonal
  // x, y, z coordinates
  template <class T>
  std::array<T, 3> fractionalize(const std::array<T, 3>& coord) const;

  // Returns the cartesian coordinates corresponding to the supplied fractional
  // x, y, z coordinates
  template <class T>
  std::array<T, 3> orthogonalize(const std::array<T, 3>& coord) const;

  // Calculates (sin(theta) / lambda)^2 for a reflection with the given indices
  template <class T>
  double stolsq(T h, T k, T l) const;

  // Calculates the resolution in Angstroms for a reflection with the given
  // indices
  template <class T>
  double resolution(T h, T k, T l) const;

  // Returns the space group associated with this unit_cell
  const spacegroup& space_group() const { return *spcgrp; };

private:
  // stored as a pointer rather than reference to allow implicitly defined copy
  // operations
  spacegroup* spcgrp;

  // real-space cell constants
  double a;
  double b;
  double c;
  double alph;
  double bet;
  double gam;
  double vol;

  // reciprocal-space cell constants
  double ra;
  double rb;
  double rc;
  double ralpha;
  double rbeta;
  double rgamma;

  // these values are stored for quick calculation of resolution
  double arbrgr;
  double arcrbr;
  double brcrar;
  double arar;
  double brbr;
  double crcr;

  // orthogonalization and fractionalization matrices
  std::array<std::array<double, 3>, 3> frac_to_cart;
  std::array<std::array<double, 3>, 3> cart_to_frac;
};

template <class T>
double unit_cell::stolsq(T h, T k, T l) const {
  return (h * h * arar) + (k * k * brbr) + (l * l * crcr) + (h * k * arbrgr) +
         (h * l * arcrbr) + (k * l * brcrar);
}

template <class T>
double unit_cell::resolution(T h, T k, T l) const {
  auto stol = std::sqrt(stolsq(h, k, l));
  // if the sin(theta) / lambda value is approaching zero, return an appropriate
  // resolution value
  if (stol < std::numeric_limits<double>::min()) {
    return (std::numeric_limits<double>::has_infinity)
               ? std::numeric_limits<double>::infinity()
               : std::numeric_limits<double>::max();
  }
  return 0.5 / stol;
}

template <class T>
inline std::array<T, 3>
unit_cell::fractionalize(const std::array<T, 3>& coord) const {
  return {{cart_to_frac[0][0] * coord[0] + cart_to_frac[0][1] * coord[1] +
               cart_to_frac[0][2] * coord[2],
           cart_to_frac[1][1] * coord[1] + cart_to_frac[1][2] * coord[2],
           cart_to_frac[2][2] * coord[2]}};
}

template <class T>
inline std::array<T, 3>
unit_cell::orthogonalize(const std::array<T, 3>& coord) const {
  return {{frac_to_cart[0][0] * coord[0] + frac_to_cart[0][1] * coord[1] +
               frac_to_cart[0][2] * coord[2],
           frac_to_cart[1][1] * coord[1] + frac_to_cart[1][2] * coord[2],
           frac_to_cart[2][2] * coord[2]}};
}

// Algorithms are from Stout & Jensen, "X-ray Structure Determination - A
// Practical Guide",
// 2nd edition, 1989.

inline unit_cell::unit_cell(int space_group_number, double a_axis_length,
                            double b_axis_length, double c_axis_length,
                            double alpha_angle, double beta_angle,
                            double gamma_angle)
    : spcgrp(&spacegroup::from_number(space_group_number)), a(a_axis_length),
      b(b_axis_length), c(c_axis_length), alph(alpha_angle), bet(beta_angle),
      gam(gamma_angle) {

  // if greater than pi, assume the angles were specified in degrees, and
  // convert
  const auto Deg2Rad = geom::PI / 180.0;
  if (alph > geom::PI)
    alph *= Deg2Rad;
  if (bet > geom::PI)
    bet *= Deg2Rad;
  if (gam > geom::PI)
    gam *= Deg2Rad;
  auto sa = std::sin(alph);
  auto sb = std::sin(bet);
  auto sg = std::sin(gam);
  auto ca = std::cos(alph);
  auto cb = std::cos(bet);
  auto cg = std::cos(gam);
  vol = a * b * c *
        std::sqrt(1.0 - ca * ca - cb * cb - cg * cg + 2.0 * ca * cb * cg);

  ra = b * c * sa / vol;
  rb = a * c * sb / vol;
  rc = a * b * sg / vol;

  auto car = (cb * cg - ca) / (sb * sg);
  auto cbr = (ca * cg - cb) / (sa * sg);
  auto cgr = (ca * cb - cg) / (sa * sb);
  ralpha = std::acos(car);
  rbeta = std::acos(cbr);
  rgamma = std::acos(cgr);

  // calculate intermediate values for resolution calculations
  arbrgr = ra * rb * cgr / 2.0;
  arcrbr = ra * rc * cbr / 2.0;
  brcrar = rb * rc * car / 2.0;
  arar = ra * ra / 4.0;
  brbr = rb * rb / 4.0;
  crcr = rc * rc / 4.0;

  frac_to_cart = {{{{a, b * cg, c * cb}},
                   {{0.0, b * sg, -c * sb * car}},
                   {{0.0, 0.0, c * sb * std::sin(ralpha)}}}};
  geom::invert(frac_to_cart, cart_to_frac);
}
}

#endif
